"""
A tool with an initial set of error analysis for in-order parsing.

Analyzes the first error created in the parser

TODO: there are more errors to analyze, and see below for a case where attachment is misidentified as bracket
"""

from enum import Enum

from stanza.models.constituency.dynamic_oracle import advance_past_constituents
from stanza.models.constituency.parse_transitions import Shift, CompoundUnary, OpenConstituent, CloseConstituent, TransitionScheme, Finalize
from stanza.models.constituency.transition_sequence import build_sequence

class FirstError(Enum):
    NONE                        = 1
    UNKNOWN                     = 2
    WRONG_OPEN_LABEL_NO_CASCADE = 3
    WRONG_OPEN_LABEL_CASCADE    = 4
    WRONG_SUBTREE_NO_CASCADE    = 5
    WRONG_SUBTREE_CASCADE       = 6
    EXTRA_ATTACHMENT            = 7
    MISSING_ATTACHMENT          = 8
    EXTRA_BRACKET_NO_CASCADE    = 9
    EXTRA_BRACKET_CASCADE       = 10
    MISSING_BRACKET_NO_CASCADE  = 11
    MISSING_BRACKET_CASCADE     = 12

def advance_past_unaries(sequence, idx):
    while idx + 2 < len(sequence) and isinstance(sequence[idx+1], OpenConstituent) and isinstance(sequence[idx+2], CloseConstituent):
        idx += 2
    return idx

def check_attachment_error(gold_sequence, pred_sequence, idx, error_type):
    # this will find the Close that closes the constituent that
    # was just closed in the gold sequence
    # hopefully we will have built the same constituent(s)
    # that were built after the gold sequence closed
    pred_close_idx = advance_past_constituents(pred_sequence, idx)
    gold_close_idx = pred_close_idx + 1
    #gold_close_idx = find_in_order_constituent_end(gold_sequence, idx+1) # +1 represents, start counting from the Shift
    #pred_close_idx = find_in_order_constituent_end(pred_sequence, idx)
    if gold_sequence[idx+1:gold_close_idx] != pred_sequence[idx:pred_close_idx]:
        return FirstError.UNKNOWN
    if (isinstance(gold_sequence[gold_close_idx], CloseConstituent) and
        isinstance(pred_sequence[pred_close_idx], CloseConstituent) and
        isinstance(pred_sequence[pred_close_idx+1], CloseConstituent)):
        #print(gold_sequence)
        #print(gold_close_idx)
        #print(pred_sequence)
        #print(pred_close_idx)
        #print("{:P}".format(gold_tree))
        #print("{:P}".format(pred_tree))
        #print("=================")
        return error_type

    return None

def analyze_tree(gold_tree, pred_tree):
    if gold_tree == pred_tree:
        return FirstError.NONE

    gold_sequence = build_sequence(gold_tree, TransitionScheme.IN_ORDER)
    pred_sequence = build_sequence(pred_tree, TransitionScheme.IN_ORDER)

    for idx, (gold_trans, pred_trans) in enumerate(zip(gold_sequence, pred_sequence)):
        if gold_trans != pred_trans:
            break
    else:
        # guess only the tags were different?
        return FirstError.NONE

    if isinstance(gold_trans, CloseConstituent) and isinstance(pred_trans, Shift) and isinstance(gold_sequence[idx + 1], Shift):
        # perhaps this is an attachment error
        # we can see if the exact same sequence of moved constituent was built
        error = check_attachment_error(gold_sequence, pred_sequence, idx, FirstError.EXTRA_ATTACHMENT)
        if error is not None:
            return error

    if isinstance(pred_trans, CloseConstituent) and isinstance(gold_trans, Shift) and isinstance(pred_sequence[idx + 1], Shift):
        # perhaps this is an attachment error
        # we can see if the exact same sequence of moved constituent was built
        error = check_attachment_error(pred_sequence, gold_sequence, idx, FirstError.MISSING_ATTACHMENT)
        if error is not None:
            return error

    if isinstance(gold_trans, OpenConstituent) and isinstance(pred_trans, OpenConstituent):
        gold_close_idx = advance_past_constituents(gold_sequence, idx+1)
        gold_unary_idx = advance_past_unaries(gold_sequence, gold_close_idx)

        pred_close_idx = advance_past_constituents(pred_sequence, idx+1)
        pred_unary_idx = advance_past_unaries(pred_sequence, pred_close_idx)
        if gold_sequence[idx+1:gold_close_idx] != pred_sequence[idx+1:pred_close_idx]:
            # maybe the internal structure is the same?
            # actually, if the number of shifts inside is the same,
            # then the words shifted were the same,
            # so the internal structure is different but the parser
            # is getting back on track after closing
            if (sum(isinstance(gt, Shift) for gt in gold_sequence[idx+1:gold_close_idx]) ==
                sum(isinstance(pt, Shift) for pt in pred_sequence[idx+1:pred_close_idx])):
                if gold_sequence[gold_unary_idx:] == pred_sequence[pred_unary_idx:]:
                    return FirstError.WRONG_SUBTREE_NO_CASCADE
                else:
                    return FirstError.WRONG_SUBTREE_CASCADE
            return FirstError.UNKNOWN
        # at this point, everything is the same aside from the open being a different label
        if gold_sequence[gold_unary_idx:] == pred_sequence[pred_unary_idx:]:
            return FirstError.WRONG_OPEN_LABEL_NO_CASCADE
        else:
            return FirstError.WRONG_OPEN_LABEL_CASCADE

    if isinstance(gold_trans, Shift) and isinstance(pred_trans, OpenConstituent):
        # This could be a case of an extra bracket inserted into the tree
        # We will search for the end of the new bracket, then check if
        # all the children were properly constructed the way the gold sequence wanted to,
        # aside from the extra bracket

        # TODO: this is also capturing what are effectively attachment
        # errors in the case of nested nodes (S over S) where a node
        # at the start should have been connected to the below node
        #   gold:
        #  (ROOT
        #    (S
        #      (S
        #        (`` ``)
        #        (NP (PRP$ Our) (NN balance) (NNS sheets))
        #        (VP
        #          (VBP look)
        #          (SBAR
        #            (IN like)
        #            (S
        #              (NP (PRP they))
        #              (VP
        #                (VBD came)
        #                (PP
        #                  (IN from)
        #                  (NP
        #                    (NP (NNP Alice) (POS 's))
        #                    (NN wonderland)))))))
        #        (, ,)
        #        ('' ''))
        #      (NP (NNP Mr.) (NNP Fromstein))
        #      (VP (VBD said))
        #      (. .)))
        #
        #  pred:
        #  (ROOT
        #    (S
        #      (`` ``)
        #      (S
        #        (NP (PRP$ Our) (NN balance) (NNS sheets))
        #        (VP
        #          (VBP look)
        #          (SBAR
        #            (IN like)
        #            (S
        #              (NP (PRP they))
        #              (VP
        #                (VBD came)
        #                (PP
        #                  (IN from)
        #                  (NP
        #                    (NP (NNP Alice) (POS 's))
        #                    (NN wonderland))))))))
        #      (, ,)
        #      ('' '')
        #      (NP (NNP Mr.) (NNP Fromstein))
        #      (VP (VBD said))
        #      (. .)))

        pred_close_idx = advance_past_constituents(pred_sequence, idx+1)
        pred_unary_idx = advance_past_unaries(pred_sequence, pred_close_idx + 1)
        if gold_sequence[idx:pred_close_idx-1] == pred_sequence[idx+1:pred_close_idx]:
            #print(gold_sequence)
            #print(pred_sequence)
            #print(idx, pred_close_idx)
            #print("{:P}".format(gold_tree))
            #print("{:P}".format(pred_tree))
            #print("=================")
            gold_unary_idx = advance_past_unaries(gold_sequence, pred_close_idx - 1)
            if pred_sequence[pred_unary_idx:] == gold_sequence[gold_unary_idx:]:
                return FirstError.EXTRA_BRACKET_NO_CASCADE
            else:
                return FirstError.EXTRA_BRACKET_CASCADE

    if isinstance(pred_trans, Shift) and isinstance(gold_trans, OpenConstituent):
        # presumably this has attachment errors as well, similarly to EXTRA_BRACKET
        gold_close_idx = advance_past_constituents(gold_sequence, idx+1)
        gold_unary_idx = advance_past_unaries(gold_sequence, gold_close_idx + 1)
        if pred_sequence[idx:gold_close_idx-1] == gold_sequence[idx+1:gold_close_idx]:
            #print(gold_sequence)
            #print(pred_sequence)
            #print(idx, gold_close_idx)
            #print("{:P}".format(gold_tree))
            #print("{:P}".format(pred_tree))
            #print("=================")
            pred_unary_idx = advance_past_unaries(pred_sequence, gold_close_idx - 1)
            if pred_sequence[pred_unary_idx:] == gold_sequence[gold_unary_idx:]:
                return FirstError.MISSING_BRACKET_NO_CASCADE
            else:
                return FirstError.MISSING_BRACKET_CASCADE

    return FirstError.UNKNOWN



import pytest
import stanza

from stanza.tests import *
from stanza.models.common.data import get_augment_ratio, augment_punct

pytestmark = [pytest.mark.travis, pytest.mark.pipeline]

def test_augment_ratio():
    data = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    should_augment = lambda x: x >= 3
    can_augment = lambda x: x >= 4
    # check that zero is returned if no augmentation is needed
    # which will be the case since 2 are already satisfactory
    assert get_augment_ratio(data, should_augment, can_augment, desired_ratio=0.1) == 0.0

    # this should throw an error
    with pytest.raises(AssertionError):
        get_augment_ratio(data, can_augment, should_augment)

    # with a desired ratio of 0.4,
    # there are already 2 that don't need augmenting
    # and 7 that are eligible to be augmented
    # so 2/7 will need to be augmented
    assert get_augment_ratio(data, should_augment, can_augment, desired_ratio=0.4) == pytest.approx(2/7)

def test_augment_punct():
    data = [["Simple", "test", "."]]
    should_augment = lambda x: x[-1] == "."
    can_augment = should_augment
    new_data = augment_punct(data, 1.0, should_augment, can_augment)
    assert new_data == [["Simple", "test"]]



import random
import logging
import copy
import torch
from collections import namedtuple

from torch.utils.data import DataLoader as DL
from torch.utils.data.sampler import Sampler
from torch.nn.utils.rnn import pad_sequence

from stanza.models.common.bert_embedding import filter_data, needs_length_filter
from stanza.models.common.data import map_to_ids, get_long_tensor, get_float_tensor, sort_all
from stanza.models.common.vocab import PAD_ID, VOCAB_PREFIX, CharVocab
from stanza.models.pos.vocab import WordVocab, XPOSVocab, FeatureVocab, MultiVocab
from stanza.models.pos.xpos_vocab_factory import xpos_vocab_factory
from stanza.models.common.doc import *

logger = logging.getLogger('stanza')

DataSample = namedtuple("DataSample", "word char upos xpos feats pretrain text")
DataBatch = namedtuple("DataBatch", "words words_mask wordchars wordchars_mask upos xpos ufeats pretrained orig_idx word_orig_idx lens word_lens text idx")

class Dataset:
    def __init__(self, doc, args, pretrain, vocab=None, evaluation=False, sort_during_eval=False, bert_tokenizer=None, **kwargs):
        self.args = args
        self.eval = evaluation
        self.shuffled = not self.eval
        self.sort_during_eval = sort_during_eval
        self.doc = doc

        if vocab is None:
            self.vocab = Dataset.init_vocab([doc], args)
        else:
            self.vocab = vocab

        self.has_upos = not all(x is None or x == '_' for x in doc.get(UPOS, as_sentences=False))
        self.has_xpos = not all(x is None or x == '_' for x in doc.get(XPOS, as_sentences=False))
        self.has_feats = not all(x is None or x == '_' for x in doc.get(FEATS, as_sentences=False))

        data = self.load_doc(self.doc)
        # filter out the long sentences if bert is used
        if self.args.get('bert_model', None) and needs_length_filter(self.args['bert_model']):
            data = filter_data(self.args['bert_model'], data, bert_tokenizer)

        # handle pretrain; pretrain vocab is used when args['pretrain'] == True and pretrain is not None
        self.pretrain_vocab = None
        if pretrain is not None and args['pretrain']:
            self.pretrain_vocab = pretrain.vocab

        # filter and sample data
        if args.get('sample_train', 1.0) < 1.0 and not self.eval:
            keep = int(args['sample_train'] * len(data))
            data = random.sample(data, keep)
            logger.debug("Subsample training set with rate {:g}".format(args['sample_train']))

        data = self.preprocess(data, self.vocab, self.pretrain_vocab, args)

        self.data = data

        self.num_examples = len(data)
        self.__punct_tags = self.vocab["upos"].map(["PUNCT"])
        self.augment_nopunct = self.args.get("augment_nopunct", 0.0)

    @staticmethod
    def init_vocab(docs, args):
        data = [x for doc in docs for x in Dataset.load_doc(doc)]
        charvocab = CharVocab(data, args['shorthand'])
        wordvocab = WordVocab(data, args['shorthand'], cutoff=args['word_cutoff'], lower=True)
        uposvocab = WordVocab(data, args['shorthand'], idx=1)
        xposvocab = xpos_vocab_factory(data, args['shorthand'])
        try:
            featsvocab = FeatureVocab(data, args['shorthand'], idx=3)
        except ValueError as e:
            raise ValueError("Unable to build features vocab.  Please check the Features column of your data for an error which may match the following description.") from e
        vocab = MultiVocab({'char': charvocab,
                            'word': wordvocab,
                            'upos': uposvocab,
                            'xpos': xposvocab,
                            'feats': featsvocab})
        return vocab

    def preprocess(self, data, vocab, pretrain_vocab, args):
        processed = []
        for sent in data:
            processed_sent = DataSample(
                word = [vocab['word'].map([w[0] for w in sent])],
                char = [[vocab['char'].map([x for x in w[0]]) for w in sent]],
                upos = [vocab['upos'].map([w[1] for w in sent])],
                xpos = [vocab['xpos'].map([w[2] for w in sent])],
                feats = [vocab['feats'].map([w[3] for w in sent])],
                pretrain = ([pretrain_vocab.map([w[0].lower() for w in sent])]
                            if pretrain_vocab is not None
                           else [[PAD_ID] * len(sent)]),
                text = [w[0] for w in sent]
            )
            processed.append(processed_sent)

        return processed

    def __len__(self):
        return len(self.data)

    def __mask(self, upos):
        """Returns a torch boolean about which elements should be masked out"""

        # creates all false mask
        mask = torch.zeros_like(upos, dtype=torch.bool)

        ### augmentation 1: punctuation augmentation ###
        # tags that needs to be checked, currently only PUNCT
        if random.uniform(0,1) < self.augment_nopunct:
            for i in self.__punct_tags:
                # generate a mask for the last element
                last_element = torch.zeros_like(upos, dtype=torch.bool)
                last_element[..., -1] = True
                # we or the bitmask against the existing mask
                # if it satisfies, we remove the word by masking it
                # to true
                #
                # if your input is just a lone punctuation, we perform
                # no masking
                if not torch.all(upos.eq(torch.tensor([[i]]))):
                    mask |= ((upos == i) & (last_element))

        return mask

    def __getitem__(self, key):
        """Retrieves a sample from the dataset.

        Retrieves a sample from the dataset. This function, for the
        most part, is spent performing ad-hoc data augmentation and
        restoration. It recieves a DataSample object from the storage,
        and returns an almost-identical DataSample object that may
        have been augmented with /possibly/ (depending on augment_punct
        settings) PUNCT chopped.

        **Important Note**
        ------------------
        If you would like to load the data into a model, please convert
        this Dataset object into a DataLoader via self.to_loader(). Then,
        you can use the resulting object like any other PyTorch data
        loader. As masks are calculated ad-hoc given the batch, the samples
        returned from this object doesn't have the appropriate masking.

        Motivation
        ----------
        Why is this here? Every time you call next(iter(dataloader)), it calls
        this function. Therefore, if we augmented each sample on each iteration,
        the model will see dynamically generated augmentation.
        Furthermore, PyTorch dataloader handles shuffling natively.

        Parameters
        ----------
        key : int
            the integer ID to from which to retrieve the key.

        Returns
        -------
        DataSample
            The sample of data you requested, with augmentation.
        """
        # get a sample of the input data
        sample = self.data[key]

        # some data augmentation requires constructing a mask based on upos.
        # For instance, sometimes we'd like to mask out ending sentence punctuation.
        # We copy the other items here so that any edits made because
        # of the mask don't clobber the version owned by the Dataset
        # convert to tensors
        # TODO: only store single lists per data entry?
        words = torch.tensor(sample.word[0])
        # convert the rest to tensors
        upos = torch.tensor(sample.upos[0]) if self.has_upos else None
        xpos = torch.tensor(sample.xpos[0]) if self.has_xpos else None
        ufeats = torch.tensor(sample.feats[0]) if self.has_feats else None
        pretrained = torch.tensor(sample.pretrain[0])

        # and deal with char & raw_text
        char = sample.char[0]
        raw_text = sample.text

        # some data augmentation requires constructing a mask based on
        # which upos. For instance, sometimes we'd like to mask out ending
        # sentence punctuation. The mask is True if we want to remove the element
        if self.has_upos and upos is not None and not self.eval:
            # perform actual masking
            mask = self.__mask(upos)
        else:
            # dummy mask that's all false
            mask = None
        if mask is not None:
            mask_index = mask.nonzero()

            # mask out the elements that we need to mask out
            for mask in mask_index:
                mask = mask.item()
                words[mask] = PAD_ID
                if upos is not None:
                    upos[mask] = PAD_ID
                if xpos is not None:
                    # TODO: test the multi-dimension xpos
                    xpos[mask, ...] = PAD_ID
                if ufeats is not None:
                    ufeats[mask, ...] = PAD_ID
                pretrained[mask] = PAD_ID
                char = char[:mask] + char[mask+1:]
                raw_text = raw_text[:mask] + raw_text[mask+1:]

        # get each character from the input sentnece
        # chars = [w for sent in char for w in sent]

        return DataSample(words, char, upos, xpos, ufeats, pretrained, raw_text), key

    def __iter__(self):
        for i in range(self.__len__()):
            yield self.__getitem__(i)

    def to_loader(self, **kwargs):
        """Converts self to a DataLoader """

        return DL(self,
                  collate_fn=Dataset.__collate_fn,
                  **kwargs)

    def to_length_limited_loader(self, batch_size, maximum_tokens):
        sampler = LengthLimitedBatchSampler(self, batch_size, maximum_tokens)
        return DL(self,
                  collate_fn=Dataset.__collate_fn,
                  batch_sampler = sampler)

    @staticmethod
    def __collate_fn(data):
        """Function used by DataLoader to pack data"""
        (data, idx) = zip(*data)
        (words, wordchars, upos, xpos, ufeats, pretrained, text) = zip(*data)

        # collate_fn is given a list of length batch size
        batch_size = len(data)

        # sort sentences by lens for easy RNN operations
        lens = [torch.sum(x != PAD_ID) for x in words]
        (words, wordchars, upos, xpos,
         ufeats, pretrained, text), orig_idx = sort_all((words, wordchars, upos, xpos,
                                                         ufeats, pretrained, text), lens)
        lens = [torch.sum(x != PAD_ID) for x in words] # we need to reinterpret lengths for the RNN

        # combine all words into one large list, and sort for easy charRNN ops
        wordchars = [w for sent in wordchars for w in sent]
        word_lens = [len(x) for x in wordchars]
        (wordchars,), word_orig_idx = sort_all([wordchars], word_lens)
        word_lens = [len(x) for x in wordchars] # we need to reinterpret lengths for the RNN

        # We now pad everything
        words = pad_sequence(words, True, PAD_ID)
        if None not in upos:
            upos = pad_sequence(upos, True, PAD_ID)
        else:
            upos = None
        if None not in xpos:
            xpos = pad_sequence(xpos, True, PAD_ID)
        else:
            xpos = None
        if None not in ufeats:
            ufeats = pad_sequence(ufeats, True, PAD_ID)
        else:
            ufeats = None
        pretrained = pad_sequence(pretrained, True, PAD_ID)
        wordchars = get_long_tensor(wordchars, len(word_lens))

        # and finally create masks for the padding indices
        words_mask = torch.eq(words, PAD_ID)
        wordchars_mask = torch.eq(wordchars, PAD_ID)

        return DataBatch(words, words_mask, wordchars, wordchars_mask, upos, xpos, ufeats,
                         pretrained, orig_idx, word_orig_idx, lens, word_lens, text, idx)

    @staticmethod
    def load_doc(doc):
        data = doc.get([TEXT, UPOS, XPOS, FEATS], as_sentences=True)
        data = Dataset.resolve_none(data)
        return data

    @staticmethod
    def resolve_none(data):
        # replace None to '_'
        for sent_idx in range(len(data)):
            for tok_idx in range(len(data[sent_idx])):
                for feat_idx in range(len(data[sent_idx][tok_idx])):
                    if data[sent_idx][tok_idx][feat_idx] is None:
                        data[sent_idx][tok_idx][feat_idx] = '_'
        return data

class LengthLimitedBatchSampler(Sampler):
    """
    Batches up the text in batches of batch_size, but cuts off each time a batch reaches maximum_tokens

    Intent is to avoid GPU OOM in situations where one sentence is significantly longer than expected,
    leaving a batch too large to fit in the GPU

    Sentences which are longer than maximum_tokens by themselves are put in their own batches
    """
    def __init__(self, data, batch_size, maximum_tokens):
        """
        Precalculate the batches, making it so len and iter just read off the precalculated batches
        """
        self.data = data
        self.batch_size = batch_size
        self.maximum_tokens = maximum_tokens

        self.batches = []
        current_batch = []
        current_length = 0

        for item, item_idx in data:
            item_len = len(item.word)
            if maximum_tokens and item_len > maximum_tokens:
                if len(current_batch) > 0:
                    self.batches.append(current_batch)
                    current_batch = []
                    current_length = 0
                self.batches.append([item_idx])
                continue
            if len(current_batch) + 1 > batch_size or (maximum_tokens and item_len + current_length > maximum_tokens):
                self.batches.append(current_batch)
                current_batch = []
                current_length = 0
            current_batch.append(item_idx)
            current_length += item_len

        if len(current_batch) > 0:
            self.batches.append(current_batch)

    def __len__(self):
        return len(self.batches)

    def __iter__(self):
        for batch in self.batches:
            current_batch = []
            for idx in batch:
                current_batch.append(idx)
            yield current_batch


class ShuffledDataset:
    """A wrapper around one or more datasets which shuffles the data in batch_size chunks

    This means that if multiple datasets are passed in, the batches
    from each dataset are shuffled together, with one batch being
    entirely members of the same dataset.

    The main use case of this is that in the tagger, there are cases
    where batches from different datasets will have different
    properties, such as having or not having UPOS tags.  We found that
    it is actually somewhat tricky to make the model's loss function
    (in model.py) properly represent batches with mixed w/ and w/o
    property, whereas keeping one entire batch together makes it a lot
    easier to process.

    The mechanism for the shuffling is that the iterator first makes a
    list long enough to represent each batch from each dataset,
    tracking the index of the dataset it is coming from, then shuffles
    that list.  Another alternative would be to use a weighted
    randomization approach, but this is very simple and the memory
    requirements are not too onerous.

    Note that the batch indices are wasteful in the case of only one
    underlying dataset, which is actually the most common use case,
    but the overhead is small enough that it probably isn't worth
    special casing the one dataset version.
    """
    def __init__(self, datasets, batch_size):
        self.batch_size = batch_size
        self.datasets = datasets
        self.loaders = [x.to_loader(batch_size=self.batch_size, shuffle=True) for x in self.datasets]

    def __iter__(self):
        iterators = [iter(x) for x in self.loaders]
        lengths = [len(x) for x in self.loaders]
        indices = [[x] * y for x, y in enumerate(lengths)]
        indices = [idx for inner in indices for idx in inner]
        random.shuffle(indices)

        for idx in indices:
            yield(next(iterators[idx]))

    def __len__(self):
        return sum(len(x) for x in self.datasets)



import random
import numpy as np
import os
from collections import Counter
import logging
import torch

import stanza.models.common.seq2seq_constant as constant
from stanza.models.common.data import map_to_ids, get_long_tensor, get_float_tensor, sort_all
from stanza.models.common.vocab import DeltaVocab
from stanza.models.lemma.vocab import Vocab, MultiVocab
from stanza.models.lemma import edit
from stanza.models.common.doc import *

logger = logging.getLogger('stanza')

class DataLoader:
    def __init__(self, doc, batch_size, args, vocab=None, evaluation=False, conll_only=False, skip=None, expand_unk_vocab=False):
        self.batch_size = batch_size
        self.args = args
        self.eval = evaluation
        self.shuffled = not self.eval
        self.doc = doc

        data = self.raw_data()

        if conll_only: # only load conll file
            return

        if skip is not None:
            assert len(data) == len(skip)
            data = [x for x, y in zip(data, skip) if not y]

        # handle vocab
        if vocab is not None:
            if expand_unk_vocab:
                pos_vocab = vocab['pos']
                char_vocab = DeltaVocab(data, vocab['char'])
                self.vocab = MultiVocab({'char': char_vocab, 'pos': pos_vocab})
            else:
                self.vocab = vocab
        else:
            self.vocab = dict()
            char_vocab, pos_vocab = self.init_vocab(data)
            self.vocab = MultiVocab({'char': char_vocab, 'pos': pos_vocab})

        # filter and sample data
        if args.get('sample_train', 1.0) < 1.0 and not self.eval:
            keep = int(args['sample_train'] * len(data))
            data = random.sample(data, keep)
            logger.debug("Subsample training set with rate {:g}".format(args['sample_train']))

        data = self.preprocess(data, self.vocab['char'], self.vocab['pos'], args)
        # shuffle for training
        if self.shuffled:
            indices = list(range(len(data)))
            random.shuffle(indices)
            data = [data[i] for i in indices]
        self.num_examples = len(data)

        # chunk into batches
        data = [data[i:i+batch_size] for i in range(0, len(data), batch_size)]
        self.data = data
        logger.debug("{} batches created.".format(len(data)))

    def init_vocab(self, data):
        assert self.eval is False, "Vocab file must exist for evaluation"
        char_data = "".join(d[0] + d[2] for d in data)
        char_vocab = Vocab(char_data, self.args['lang'])
        pos_data = [d[1] for d in data]
        pos_vocab = Vocab(pos_data, self.args['lang'])
        return char_vocab, pos_vocab

    def preprocess(self, data, char_vocab, pos_vocab, args):
        processed = []
        for d in data:
            edit_type = edit.EDIT_TO_ID[edit.get_edit_type(d[0], d[2])]
            src = list(d[0])
            src = [constant.SOS] + src + [constant.EOS]
            src = char_vocab.map(src)
            pos = d[1]
            pos = pos_vocab.unit2id(pos)
            tgt = list(d[2])
            tgt_in = char_vocab.map([constant.SOS] + tgt)
            tgt_out = char_vocab.map(tgt + [constant.EOS])
            processed += [[src, tgt_in, tgt_out, pos, edit_type, d[0]]]
        return processed

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        """ Get a batch with index. """
        if not isinstance(key, int):
            raise TypeError
        if key < 0 or key >= len(self.data):
            raise IndexError
        batch = self.data[key]
        batch_size = len(batch)
        batch = list(zip(*batch))
        assert len(batch) == 6

        # sort all fields by lens for easy RNN operations
        lens = [len(x) for x in batch[0]]
        batch, orig_idx = sort_all(batch, lens)

        # convert to tensors
        src = batch[0]
        src = get_long_tensor(src, batch_size)
        src_mask = torch.eq(src, constant.PAD_ID)
        tgt_in = get_long_tensor(batch[1], batch_size)
        tgt_out = get_long_tensor(batch[2], batch_size)
        pos = torch.LongTensor(batch[3])
        edits = torch.LongTensor(batch[4])
        text = batch[5]
        assert tgt_in.size(1) == tgt_out.size(1), "Target input and output sequence sizes do not match."
        return src, src_mask, tgt_in, tgt_out, pos, edits, orig_idx, text

    def __iter__(self):
        for i in range(self.__len__()):
            yield self.__getitem__(i)

    def raw_data(self):
        return self.load_doc(self.doc, self.args.get('caseless', False), self.args.get('skip_blank_lemmas', False), self.eval)

    @staticmethod
    def load_doc(doc, caseless, skip_blank_lemmas, evaluation):
        if evaluation:
            data = doc.get([TEXT, UPOS, LEMMA])
        else:
            data = doc.get([TEXT, UPOS, LEMMA, HEAD, DEPREL, MISC], as_sentences=True)
            data = DataLoader.remove_goeswith(data)
            data = DataLoader.extract_correct_forms(data)
        data = DataLoader.resolve_none(data)
        if not evaluation and skip_blank_lemmas:
            data = DataLoader.skip_blank_lemmas(data)
        if caseless:
            data = DataLoader.lowercase_data(data)
        return data

    @staticmethod
    def extract_correct_forms(data):
        """
        Here we go through the raw data and use the CorrectForm of words tagged with CorrectForm

        In addition, if the incorrect form of the word is not present in the training data,
        we keep the incorrect form for the lemmatizer to learn from.
        This way, it can occasionally get things right in misspelled input text.

        We do check for and eliminate words where the incorrect form is already known as the
        lemma for a different word.  For example, in the English datasets, there is a "busy"
        which was meant to be "buys", and we don't want the model to learn to lemmatize "busy" to "buy"
        """
        new_data = []
        incorrect_forms = []
        for word in data:
            misc = word[-1]
            if not misc:
                new_data.append(word[:3])
                continue
            misc = misc.split("|")
            for piece in misc:
                if piece.startswith("CorrectForm="):
                    cf = piece.split("=", maxsplit=1)[1]
                    # treat the CorrectForm as the desired word
                    new_data.append((cf, word[1], word[2]))
                    # and save the broken one for later in case it wasn't used anywhere else
                    incorrect_forms.append((cf, word))
                    break
            else:
                # if no CorrectForm, just keep the word as normal
                new_data.append(word[:3])
        known_words = {x[0] for x in new_data}
        for correct_form, word in incorrect_forms:
            if word[0] not in known_words:
                new_data.append(word[:3])
        return new_data

    @staticmethod
    def remove_goeswith(data):
        """
        This method specifically removes words that goeswith something else, along with the something else

        The purpose is to eliminate text such as

1	Ken	kenrice@enroncommunications	X	GW	Typo=Yes	0	root	0:root	_
2	Rice@ENRON	_	X	GW	_	1	goeswith	1:goeswith	_
3	COMMUNICATIONS	_	X	ADD	_	1	goeswith	1:goeswith	_
        """
        filtered_data = []
        remove_indices = set()
        for sentence in data:
            remove_indices.clear()
            for word_idx, word in enumerate(sentence):
                if word[4] == 'goeswith':
                    remove_indices.add(word_idx)
                    remove_indices.add(word[3]-1)
            filtered_data.extend([x for idx, x in enumerate(sentence) if idx not in remove_indices])
        return filtered_data

    @staticmethod
    def lowercase_data(data):
        for token in data:
            token[0] = token[0].lower()
        return data

    @staticmethod
    def skip_blank_lemmas(data):
        data = [x for x in data if x[2] != '_']
        return data

    @staticmethod
    def resolve_none(data):
        # replace None to '_'
        for tok_idx in range(len(data)):
            for feat_idx in range(len(data[tok_idx])):
                if data[tok_idx][feat_idx] is None:
                    data[tok_idx][feat_idx] = '_'
        return data



"""
Common utilities for Stanza resources.
"""

from collections import defaultdict, namedtuple
import errno
import hashlib
import json
import logging
import os
from pathlib import Path
import requests
import shutil
import tempfile
import zipfile

from tqdm.auto import tqdm

from stanza.utils.helper_func import make_table
from stanza.pipeline._constants import TOKENIZE, MWT, POS, LEMMA, DEPPARSE, NER, SENTIMENT
from stanza.pipeline.registry import PIPELINE_NAMES, PROCESSOR_VARIANTS
from stanza.resources.default_packages import PACKAGES
from stanza._version import __resources_version__

logger = logging.getLogger('stanza')

# set home dir for default
HOME_DIR = str(Path.home())
STANFORDNLP_RESOURCES_URL = 'https://nlp.stanford.edu/software/stanza/stanza-resources/'
STANZA_RESOURCES_GITHUB = 'https://raw.githubusercontent.com/stanfordnlp/stanza-resources/'
DEFAULT_RESOURCES_URL = os.getenv('STANZA_RESOURCES_URL', STANZA_RESOURCES_GITHUB + 'main')
DEFAULT_RESOURCES_VERSION = os.getenv(
    'STANZA_RESOURCES_VERSION',
    __resources_version__
)
DEFAULT_MODEL_URL = os.getenv('STANZA_MODEL_URL', 'default')
DEFAULT_MODEL_DIR = os.getenv(
    'STANZA_RESOURCES_DIR',
    os.path.join(HOME_DIR, 'stanza_resources')
)

PRETRAIN_NAMES = ("pretrain", "forward_charlm", "backward_charlm")

class ResourcesFileNotFoundError(FileNotFoundError):
    def __init__(self, resources_filepath):
        super().__init__(f"Resources file not found at: {resources_filepath}  Try to download the model again.")
        self.resources_filepath = resources_filepath

class UnknownLanguageError(ValueError):
    def __init__(self, unknown):
        super().__init__(f"Unknown language requested: {unknown}")
        self.unknown_language = unknown

class UnknownProcessorError(ValueError):
    def __init__(self, unknown):
        super().__init__(f"Unknown processor type requested: {unknown}")
        self.unknown_processor = unknown

ModelSpecification = namedtuple('ModelSpecification', ['processor', 'package', 'dependencies'])

def ensure_dir(path):
    """
    Create dir in case it does not exist.
    """
    Path(path).mkdir(parents=True, exist_ok=True)

def get_md5(path):
    """
    Get the MD5 value of a path.
    """
    try:
        with open(path, 'rb') as fin:
            data = fin.read()
    except OSError as e:
        if not e.filename:
            e.filename = path
        raise
    return hashlib.md5(data).hexdigest()

def unzip(path, filename):
    """
    Fully unzip a file `filename` that's in a directory `dir`.
    """
    logger.debug(f'Unzip: {path}/{filename}...')
    with zipfile.ZipFile(os.path.join(path, filename)) as f:
        f.extractall(path)

def get_root_from_zipfile(filename):
    """
    Get the root directory from a archived zip file.
    """
    zf = zipfile.ZipFile(filename, "r")
    assert len(zf.filelist) > 0, \
        f"Zip file at f{filename} seems to be corrupted. Please check it."
    return os.path.dirname(zf.filelist[0].filename)

def file_exists(path, md5):
    """
    Check if the file at `path` exists and match the provided md5 value.
    """
    return os.path.exists(path) and get_md5(path) == md5

def assert_file_exists(path, md5=None, alternate_md5=None):
    if not os.path.exists(path):
        raise FileNotFoundError(errno.ENOENT, "Cannot find expected file", path)
    if md5:
        file_md5 = get_md5(path)
        if file_md5 != md5:
            if file_md5 == alternate_md5:
                logger.debug("Found a possibly older version of file %s, md5 %s instead of %s", path, alternate_md5, md5)
            else:
                raise ValueError("md5 for %s is %s, expected %s" % (path, file_md5, md5))

def download_file(url, path, proxies, raise_for_status=False):
    """
    Download a URL into a file as specified by `path`.
    """
    verbose = logger.level in [0, 10, 20]
    r = requests.get(url, stream=True, proxies=proxies)
    if raise_for_status:
        r.raise_for_status()
    with open(path, 'wb') as f:
        file_size = r.headers.get('content-length', None)
        if file_size:
            file_size = int(file_size)
        default_chunk_size = 131072
        desc = 'Downloading ' + url
        with tqdm(total=file_size, unit='B', unit_scale=True, \
            disable=not verbose, desc=desc) as pbar:
            for chunk in r.iter_content(chunk_size=default_chunk_size):
                if chunk:
                    f.write(chunk)
                    f.flush()
                    pbar.update(len(chunk))
    return r.status_code

def request_file(url, path, proxies=None, md5=None, raise_for_status=False, log_info=True, alternate_md5=None):
    """
    A complete wrapper over download_file() that also make sure the directory of
    `path` exists, and that a file matching the md5 value does not exist.

    alternate_md5 allows for an alternate md5 that is acceptable (such as if an older version of a file is okay)
    """
    basedir = Path(path).parent
    ensure_dir(basedir)
    if file_exists(path, md5):
        if log_info:
            logger.info(f'File exists: {path}')
        else:
            logger.debug(f'File exists: {path}')
        return
    # We write data first to a temporary directory,
    # then use os.replace() so that multiple processes
    # running at the same time don't clobber each other
    # with partially downloaded files
    # This was especially common with resources.json
    with tempfile.TemporaryDirectory(dir=basedir) as temp:
        temppath = os.path.join(temp, os.path.split(path)[-1])
        download_file(url, temppath, proxies, raise_for_status)
        os.replace(temppath, path)
    assert_file_exists(path, md5, alternate_md5)
    if log_info:
        logger.info(f'Downloaded file to {path}')
    else:
        logger.debug(f'Downloaded file to {path}')

def sort_processors(processor_list):
    sorted_list = []
    for processor in PIPELINE_NAMES:
        for item in processor_list:
            if item[0] == processor:
                sorted_list.append(item)
    # going just by processors in PIPELINE_NAMES, this drops any names
    # which are not an official processor but might still be useful
    # check the list and append them to the end
    # this is especially useful when downloading pretrain or charlm models
    for processor in processor_list:
        for item in sorted_list:
            if processor[0] == item[0]:
                break
        else:
            sorted_list.append(item)
    return sorted_list

def add_mwt(processors, resources, lang):
    """Add mwt if tokenize is passed without mwt.

    If tokenize is in the list, but mwt is not, and there is a corresponding
    tokenize and mwt pair in the resources file, mwt is added so no missing
    mwt errors are raised.

    TODO: how does this handle EWT in English?
    """
    value = processors[TOKENIZE]
    if value in resources[lang][PACKAGES] and MWT in resources[lang][PACKAGES][value]:
        logger.warning("Language %s package %s expects mwt, which has been added", lang, value)
        processors[MWT] = value
    elif (value in resources[lang][TOKENIZE] and MWT in resources[lang] and value in resources[lang][MWT]):
        logger.warning("Language %s package %s expects mwt, which has been added", lang, value)
        processors[MWT] = value

def maintain_processor_list(resources, lang, package, processors, allow_pretrain=False, maybe_add_mwt=True):
    """
    Given a parsed resources file, language, and possible package
    and/or processors, expands the package to the list of processors

    Returns a list of processors
    Each item in the list of processors is a pair:
      name, then a list of ModelSpecification
    so, for example:
      [['pos', [ModelSpecification(processor='pos', package='gsd', dependencies=None)]],
       ['depparse', [ModelSpecification(processor='depparse', package='gsd', dependencies=None)]]]
    """
    processor_list = defaultdict(list)
    # resolve processor models
    if processors:
        logger.debug(f'Processing parameter "processors"...')
        if maybe_add_mwt and TOKENIZE in processors and MWT not in processors:
            add_mwt(processors, resources, lang)
        for key, plist in processors.items():
            if not isinstance(key, str):
                raise ValueError("Processor names must be strings")
            if not isinstance(plist, (tuple, list, str)):
                raise ValueError("Processor values must be strings")
            if isinstance(plist, str):
                plist = [plist]
            if key not in PIPELINE_NAMES:
                if not allow_pretrain or key not in PRETRAIN_NAMES:
                    raise UnknownProcessorError(key)
            for value in plist:
                # check if keys and values can be found
                if key in resources[lang] and value in resources[lang][key]:
                    logger.debug(f'Found {key}: {value}.')
                    processor_list[key].append(value)
                # allow values to be default in some cases
                elif value in resources[lang][PACKAGES] and key in resources[lang][PACKAGES][value]:
                    logger.debug(
                        f'Found {key}: {resources[lang][PACKAGES][value][key]}.'
                    )
                    processor_list[key].append(resources[lang][PACKAGES][value][key])
                # optional defaults will be activated if specifically turned on
                elif value in resources[lang][PACKAGES] and 'optional' in resources[lang][PACKAGES][value] and key in resources[lang][PACKAGES][value]['optional']:
                    logger.debug(
                        f"Found {key}: {resources[lang][PACKAGES][value]['optional'][key]}."
                    )
                    processor_list[key].append(resources[lang][PACKAGES][value]['optional'][key])
                # allow processors to be set to variants that we didn't implement
                elif value in PROCESSOR_VARIANTS[key]:
                    logger.debug(
                        f'Found {key}: {value}. '
                        f'Using external {value} variant for the {key} processor.'
                    )
                    processor_list[key].append(value)
                # allow lemma to be set to "identity"
                elif key == LEMMA and value == 'identity':
                    logger.debug(
                        f'Found {key}: {value}. Using identity lemmatizer.'
                    )
                    processor_list[key].append(value)
                # not a processor in the officially supported processor list
                elif key not in resources[lang]:
                    logger.debug(
                        f'{key}: {value} is not officially supported by Stanza, '
                        f'loading it anyway.'
                    )
                    processor_list[key].append(value)
                # cannot find the package for a processor and warn user
                else:
                    logger.warning(
                        f'Can not find {key}: {value} from official model list. '
                        f'Ignoring it.'
                    )
    # resolve package
    if package:
        logger.debug(f'Processing parameter "package"...')
        if PACKAGES in resources[lang] and package in resources[lang][PACKAGES]:
            for key, value in resources[lang][PACKAGES][package].items():
                if key != 'optional' and key not in processor_list:
                    logger.debug(f'Found {key}: {value}.')
                    processor_list[key].append(value)
        else:
            flag = False
            for key in PIPELINE_NAMES:
                if key not in resources[lang]: continue
                if package in resources[lang][key]:
                    flag = True
                    if key not in processor_list:
                        logger.debug(f'Found {key}: {package}.')
                        processor_list[key].append(package)
                    else:
                        logger.debug(
                            f'{key}: {package} is overwritten by '
                            f'{key}: {processors[key]}.'
                        )
            if not flag: logger.warning((f'Can not find package: {package}.'))
    processor_list = [[key, [ModelSpecification(processor=key, package=value, dependencies=None) for value in plist]] for key, plist in processor_list.items()]
    processor_list = sort_processors(processor_list)
    return processor_list

def add_dependencies(resources, lang, processor_list):
    """
    Expand the processor_list as given in maintain_processor_list to have the dependencies

    Still a list of model types to ModelSpecifications
    the dependencies are tuples: name and package
    for example:
    [['pos', (ModelSpecification(processor='pos', package='gsd', dependencies=(('pretrain', 'gsd'),)),)],
     ['depparse', (ModelSpecification(processor='depparse', package='gsd', dependencies=(('pretrain', 'gsd'),)),)]]
    """
    lang_resources = resources[lang]
    for item in processor_list:
        processor, model_specs = item
        new_model_specs = []
        for model_spec in model_specs:
            # skip dependency checking for external variants of processors and identity lemmatizer
            if not any([
                    model_spec.package in PROCESSOR_VARIANTS[processor],
                    processor == LEMMA and model_spec.package == 'identity'
                ]):
                dependencies = lang_resources.get(processor, {}).get(model_spec.package, {}).get('dependencies', [])
                dependencies = [(dependency['model'], dependency['package']) for dependency in dependencies]
                model_spec = model_spec._replace(dependencies=tuple(dependencies))
                logger.debug("Found dependencies %s for processor %s model %s", dependencies, processor, model_spec.package)
            new_model_specs.append(model_spec)
        item[1] = tuple(new_model_specs)
    return processor_list

def flatten_processor_list(processor_list):
    """
    The flattened processor list is just a list of types & packages

    For example:
      [['pos', 'gsd'], ['depparse', 'gsd'], ['pretrain', 'gsd']]
    """
    flattened_processor_list = []
    dependencies_list = []
    for item in processor_list:
        processor, model_specs = item
        for model_spec in model_specs:
            package = model_spec.package
            dependencies = model_spec.dependencies
            flattened_processor_list.append([processor, package])
            if dependencies:
                dependencies_list += [tuple(dependency) for dependency in dependencies]
    dependencies_list = [list(item) for item in set(dependencies_list)]
    for processor, package in dependencies_list:
        logger.debug(f'Find dependency {processor}: {package}.')
    flattened_processor_list += dependencies_list
    return flattened_processor_list

def set_logging_level(logging_level, verbose):
    # Check verbose for easy logging control
    if verbose == False:
        logging_level = 'ERROR'
    elif verbose == True:
        logging_level = 'INFO'

    if logging_level is None:
        # default logging level of INFO is set in stanza.__init__
        # but the user may have set it via the logging API
        # it should NOT be 0, but let's check to be sure...
        if logger.level == 0:
            logger.setLevel('INFO')
        return logger.level

    # Set logging level
    logging_level = logging_level.upper()
    all_levels = ['DEBUG', 'INFO', 'WARNING', 'WARN', 'ERROR', 'CRITICAL', 'FATAL']
    if logging_level not in all_levels:
        raise ValueError(
            f"Unrecognized logging level for pipeline: "
            f"{logging_level}. Must be one of {', '.join(all_levels)}."
        )
    logger.setLevel(logging_level)
    return logger.level

def process_pipeline_parameters(lang, model_dir, package, processors):
    # Check parameter types and convert values to lower case
    if isinstance(lang, str):
        lang = lang.strip().lower()
    elif lang is not None:
        raise TypeError(
            f"The parameter 'lang' should be str, "
            f"but got {type(lang).__name__} instead."
        )

    if isinstance(model_dir, str):
        model_dir = model_dir.strip()
    elif model_dir is not None:
        raise TypeError(
            f"The parameter 'model_dir' should be str, "
            f"but got {type(model_dir).__name__} instead."
        )

    if isinstance(processors, (str, list, tuple)):
        # Special case: processors is str, compatible with older version
        # also allow for setting alternate packages for these processors
        # via the package argument
        if package is None:
            # each processor will be 'default' for this language
            package = defaultdict(lambda: 'default')
        elif isinstance(package, str):
            # same, but now the named package will be the default instead
            default = package
            package = defaultdict(lambda: default)
        elif isinstance(package, dict):
            # the dictionary of packages will be used to build the processors dict
            # any processor not specified in package will be 'default'
            package = defaultdict(lambda: 'default', package)
        else:
            raise TypeError(
                f"The parameter 'package' should be None, str, or dict, "
                f"but got {type(package).__name__} instead."
            )
        if isinstance(processors, str):
            processors = [x.strip().lower() for x in processors.split(",")]
        processors = {
            processor: package[processor] for processor in processors
        }
        package = None
    elif isinstance(processors, dict):
        processors = {
            k.strip().lower(): ([v_i.strip().lower() for v_i in v] if isinstance(v, (tuple, list)) else v.strip().lower())
            for k, v in processors.items()
        }
    elif processors is not None:
        raise TypeError(
            f"The parameter 'processors' should be dict or str, "
            f"but got {type(processors).__name__} instead."
        )

    if isinstance(package, str):
        package = package.strip().lower()
    elif package is not None:
        raise TypeError(
            f"The parameter 'package' should be str, or a dict if 'processors' is a str, "
            f"but got {type(package).__name__} instead."
        )

    return lang, model_dir, package, processors

def download_resources_json(model_dir=DEFAULT_MODEL_DIR,
                            resources_url=DEFAULT_RESOURCES_URL,
                            resources_branch=None,
                            resources_version=DEFAULT_RESOURCES_VERSION,
                            resources_filepath=None,
                            proxies=None):
    """
    Downloads resources.json to obtain latest packages.
    """
    if resources_url == DEFAULT_RESOURCES_URL and resources_branch is not None:
        resources_url = STANZA_RESOURCES_GITHUB + resources_branch
    # handle short name for resources urls; otherwise treat it as url
    if resources_url.lower() in ('stanford', 'stanfordnlp'):
        resources_url = STANFORDNLP_RESOURCES_URL
    resources_url = f'{resources_url}/resources_{resources_version}.json'
    logger.debug('Downloading resource file from %s', resources_url)
    if resources_filepath is None:
        resources_filepath = os.path.join(model_dir, 'resources.json')
    # make request
    request_file(
        resources_url,
        resources_filepath,
        proxies,
        raise_for_status=True
    )


def load_resources_json(model_dir=DEFAULT_MODEL_DIR, resources_filepath=None):
    """
    Unpack the resources json file from the given model_dir
    """
    if resources_filepath is None:
        resources_filepath = os.path.join(model_dir, 'resources.json')
    if not os.path.exists(resources_filepath):
        raise ResourcesFileNotFoundError(resources_filepath)
    with open(resources_filepath, encoding="utf-8") as fin:
        resources = json.load(fin)
    return resources

def get_language_resources(resources, lang):
    """
    Get the resources for a lang from an already loaded resources json, following 'alias' if needed
    """
    if lang not in resources:
        return None

    lang_resources = resources[lang]
    while 'alias' in lang_resources:
        lang = lang_resources['alias']
        lang_resources = resources[lang]

    return lang_resources

def list_available_languages(model_dir=DEFAULT_MODEL_DIR,
                             resources_url=DEFAULT_RESOURCES_URL,
                             resources_branch=None,
                             resources_version=DEFAULT_RESOURCES_VERSION,
                             proxies=None):
    """
    List the non-alias languages in the resources file
    """
    download_resources_json(model_dir, resources_url, resources_branch, resources_version, resources_filepath=None, proxies=proxies)
    resources = load_resources_json(model_dir)
    # isinstance(str) is because of fields such as "url"
    # 'alias' is because we want to skip German, alias of de, for example
    languages = [lang for lang in resources
                 if not isinstance(resources[lang], str) and 'alias' not in resources[lang]]
    languages = sorted(languages)
    return languages

def expand_model_url(resources, model_url):
    """
    Returns the url in the resources dict if model_url is default, or returns the model_url
    """
    return resources['url'] if model_url.lower() == 'default' else model_url

def download_models(download_list,
                    resources,
                    lang,
                    model_dir=DEFAULT_MODEL_DIR,
                    resources_version=DEFAULT_RESOURCES_VERSION,
                    model_url=DEFAULT_MODEL_URL,
                    proxies=None,
                    log_info=True):
    lang_name = resources.get(lang, {}).get('lang_name', lang)
    download_table = make_table(['Processor', 'Package'], download_list)
    if log_info:
        log_msg = logger.info
    else:
        log_msg = logger.debug
    log_msg(
        f'Downloading these customized packages for language: '
        f'{lang} ({lang_name})...\n{download_table}'
    )

    url = expand_model_url(resources, model_url)

    # Download packages
    for key, value in download_list:
        try:
            request_file(
                url.format(resources_version=resources_version, lang=lang, filename=f"{key}/{value}.pt"),
                os.path.join(model_dir, lang, key, f'{value}.pt'),
                proxies,
                md5=resources[lang][key][value]['md5'],
                log_info=log_info,
                alternate_md5=resources[lang][key][value].get('alternate_md5', None)
            )
        except KeyError as e:
            raise ValueError(
                f'Cannot find the following processor and model name combination: '
                f'{key}, {value}. Please check if you have provided the correct model name.'
            ) from e

# main download function
def download(
        lang='en',
        model_dir=DEFAULT_MODEL_DIR,
        package='default',
        processors={},
        logging_level=None,
        verbose=None,
        resources_url=DEFAULT_RESOURCES_URL,
        resources_branch=None,
        resources_version=DEFAULT_RESOURCES_VERSION,
        model_url=DEFAULT_MODEL_URL,
        proxies=None,
        download_json=True
    ):
    # set global logging level
    set_logging_level(logging_level, verbose)
    # process different pipeline parameters
    lang, model_dir, package, processors = process_pipeline_parameters(
        lang, model_dir, package, processors
    )

    if download_json or not os.path.exists(os.path.join(model_dir, 'resources.json')):
        if not download_json:
            logger.warning("Asked to skip downloading resources.json, but the file does not exist.  Downloading anyway")
        download_resources_json(model_dir, resources_url, resources_branch, resources_version, resources_filepath=None, proxies=proxies)

    resources = load_resources_json(model_dir)
    if lang not in resources:
        raise UnknownLanguageError(lang)
    if 'alias' in resources[lang]:
        logger.info(f'"{lang}" is an alias for "{resources[lang]["alias"]}"')
        lang = resources[lang]['alias']
    lang_name = resources.get(lang, {}).get('lang_name', lang)
    url = expand_model_url(resources, model_url)

    # Default: download zipfile and unzip
    if package == 'default' and (processors is None or len(processors) == 0):
        logger.info(
            f'Downloading default packages for language: {lang} ({lang_name}) ...'
        )
        # want the URL to become, for example:
        # https://huggingface.co/stanfordnlp/stanza-af/resolve/v1.3.0/models/default.zip
        # so we hopefully start from
        # https://huggingface.co/stanfordnlp/stanza-{lang}/resolve/v{resources_version}/models/{filename}
        request_file(
            url.format(resources_version=resources_version, lang=lang, filename="default.zip"),
            os.path.join(model_dir, lang, f'default.zip'),
            proxies,
            md5=resources[lang]['default_md5'],
        )
        unzip(os.path.join(model_dir, lang), 'default.zip')
    # Customize: maintain download list
    else:
        download_list = maintain_processor_list(resources, lang, package, processors, allow_pretrain=True)
        download_list = add_dependencies(resources, lang, download_list)
        download_list = flatten_processor_list(download_list)
        download_models(download_list=download_list,
                        resources=resources,
                        lang=lang,
                        model_dir=model_dir,
                        resources_version=resources_version,
                        model_url=model_url,
                        proxies=proxies,
                        log_info=True)
    logger.info(f'Finished downloading models and saved to {model_dir}')



"""
The code in this file works to train a lemma classifier for 's
"""

import argparse
import logging
import os

import torch
import torch.nn as nn

from stanza.models.common.foundation_cache import load_pretrain
from stanza.models.lemma_classifier.base_trainer import BaseLemmaClassifierTrainer
from stanza.models.lemma_classifier.constants import DEFAULT_BATCH_SIZE
from stanza.models.lemma_classifier.lstm_model import LemmaClassifierLSTM

logger = logging.getLogger('stanza.lemmaclassifier')

class LemmaClassifierTrainer(BaseLemmaClassifierTrainer):
    """
    Class to assist with training a LemmaClassifierLSTM
    """

    def __init__(self, model_args: dict, embedding_file: str, use_charlm: bool = False, charlm_forward_file: str = None, charlm_backward_file: str = None, lr: float = 0.001, loss_func: str = None):
        """
        Initializes the LemmaClassifierTrainer class.

        Args:
            model_args (dict): Various model shape parameters
            embedding_file (str): What word embeddings file to use.  Use a Stanza pretrain .pt
            use_charlm (bool, optional): Whether to use charlm embeddings as well. Defaults to False.
            charlm_forward_file (str): Path to the forward pass embeddings for the charlm
            charlm_backward_file (str): Path to the backward pass embeddings for the charlm
            upos_emb_dim (int): The dimension size of UPOS tag embeddings
            num_heads (int): The number of attention heads to use.
            lr (float): Learning rate, defaults to 0.001.
            loss_func (str): Which loss function to use (either 'ce' or 'weighted_bce')

        Raises:
            FileNotFoundError: If the forward charlm file is not present
            FileNotFoundError: If the backward charlm file is not present
        """
        super().__init__()

        self.model_args = model_args

        # Load word embeddings
        pt = load_pretrain(embedding_file)
        self.pt_embedding = pt

        # Load CharLM embeddings
        if use_charlm and charlm_forward_file is not None and not os.path.exists(charlm_forward_file):
            raise FileNotFoundError(f"Could not find forward charlm file: {charlm_forward_file}")
        if use_charlm and charlm_backward_file is not None and not os.path.exists(charlm_backward_file):
            raise FileNotFoundError(f"Could not find backward charlm file: {charlm_backward_file}")

        # TODO: just pass around the args instead
        self.use_charlm = use_charlm
        self.charlm_forward_file = charlm_forward_file
        self.charlm_backward_file = charlm_backward_file
        self.lr = lr

        # Find loss function
        if loss_func == "ce":
            self.criterion = nn.CrossEntropyLoss()
            self.weighted_loss = False
            logger.debug("Using CE loss")
        elif loss_func == "weighted_bce":
            self.criterion = nn.BCEWithLogitsLoss()
            self.weighted_loss = True  # used to add weights during train time.
            logger.debug("Using Weighted BCE loss")
        else:
            raise ValueError("Must enter a valid loss function (e.g. 'ce' or 'weighted_bce')")

    def build_model(self, label_decoder, upos_to_id, known_words, target_words, target_upos):
        return LemmaClassifierLSTM(self.model_args, self.output_dim, self.pt_embedding, label_decoder, upos_to_id, known_words, target_words, target_upos,
                                   use_charlm=self.use_charlm, charlm_forward_file=self.charlm_forward_file, charlm_backward_file=self.charlm_backward_file)

def build_argparse():
    parser = argparse.ArgumentParser()
    parser.add_argument("--hidden_dim", type=int, default=256, help="Size of hidden layer")
    parser.add_argument('--wordvec_pretrain_file', type=str, default=os.path.join(os.path.dirname(__file__), "pretrain", "glove.pt"), help='Exact name of the pretrain file to read')
    parser.add_argument("--charlm", action='store_true', dest='use_charlm', default=False, help="Whether not to use the charlm embeddings")
    parser.add_argument('--charlm_shorthand', type=str, default=None, help="Shorthand for character-level language model training corpus.")
    parser.add_argument("--charlm_forward_file", type=str, default=os.path.join(os.path.dirname(__file__), "charlm_files", "1billion_forward.pt"), help="Path to forward charlm file")
    parser.add_argument("--charlm_backward_file", type=str, default=os.path.join(os.path.dirname(__file__), "charlm_files", "1billion_backwards.pt"), help="Path to backward charlm file")
    parser.add_argument("--upos_emb_dim", type=int, default=20, help="Dimension size for UPOS tag embeddings.")
    parser.add_argument("--use_attn", action='store_true', dest='attn', default=False, help='Whether to use multihead attention instead of LSTM.')
    parser.add_argument("--num_heads", type=int, default=0, help="Number of heads to use for multihead attention.")
    parser.add_argument("--save_name", type=str, default=os.path.join(os.path.dirname(__file__), "saved_models", "lemma_classifier_model_weighted_loss_charlm_new.pt"), help="Path to model save file")
    parser.add_argument("--lr", type=float, default=0.001, help="learning rate")
    parser.add_argument("--num_epochs", type=float, default=10, help="Number of training epochs")
    parser.add_argument("--batch_size", type=int, default=DEFAULT_BATCH_SIZE, help="Number of examples to include in each batch")
    parser.add_argument("--train_file", type=str, default=os.path.join(os.path.dirname(__file__), "data", "processed_ud_en", "combined_train.txt"), help="Full path to training file")
    parser.add_argument("--weighted_loss", action='store_true', dest='weighted_loss', default=False, help="Whether to use weighted loss during training.")
    parser.add_argument("--eval_file", type=str, default=os.path.join(os.path.dirname(__file__), "data", "processed_ud_en", "combined_dev.txt"), help="Path to dev file used to evaluate model for saves")
    parser.add_argument("--force", action='store_true', default=False, help='Whether or not to clobber an existing save file')
    return parser

def main(args=None, predefined_args=None):
    parser = build_argparse()
    args = parser.parse_args(args) if predefined_args is None else predefined_args

    wordvec_pretrain_file = args.wordvec_pretrain_file
    use_charlm = args.use_charlm
    charlm_forward_file = args.charlm_forward_file
    charlm_backward_file = args.charlm_backward_file
    upos_emb_dim = args.upos_emb_dim
    use_attention = args.attn
    num_heads = args.num_heads
    save_name = args.save_name
    lr = args.lr
    num_epochs = args.num_epochs
    train_file = args.train_file
    weighted_loss = args.weighted_loss
    eval_file = args.eval_file

    args = vars(args)

    if os.path.exists(save_name) and not args.get('force', False):
        raise FileExistsError(f"Save name {save_name} already exists. Training would override existing data. Aborting...")
    if not os.path.exists(train_file):
        raise FileNotFoundError(f"Training file {train_file} not found. Try again with a valid path.")

    logger.info("Running training script with the following args:")
    for arg in args:
        logger.info(f"{arg}: {args[arg]}")
    logger.info("------------------------------------------------------------")

    trainer = LemmaClassifierTrainer(model_args=args,
                                     embedding_file=wordvec_pretrain_file,
                                     use_charlm=use_charlm,
                                     charlm_forward_file=charlm_forward_file,
                                     charlm_backward_file=charlm_backward_file,
                                     lr=lr,
                                     loss_func="weighted_bce" if weighted_loss else "ce",
                                     )

    trainer.train(
        num_epochs=num_epochs, save_name=save_name, args=args, eval_file=eval_file, train_file=train_file
    )

    return trainer

if __name__ == "__main__":
    main()




"""
Convert the AU-KBC coreference dataset from Prof. Sobha

https://aclanthology.org/2020.wildre-1.4/

Located in /u/nlp/data/coref/tamil on the Stanford cluster
"""

import argparse
import glob
import json
from operator import itemgetter
import os
import random
import re

import stanza

from stanza.utils.datasets.coref.utils import process_document
from stanza.utils.default_paths import get_default_paths
from stanza.utils.get_tqdm import get_tqdm

tqdm = get_tqdm()

begin_re = re.compile(r"B-([0-9]+)")
in_re =  re.compile(r"I-([0-9]+)")

def write_json_file(output_filename, converted_section):
    with open(output_filename, "w", encoding="utf-8") as fout:
        json.dump(converted_section, fout, indent=2)

def read_doc(filename):
    """
    Returns the sentences and the coref markings from this filename

    sentences: a list of list of words
    corefs: a list of list of clusters, which were tagged B-num and I-num in the dataset
    """
    with open(filename, encoding="utf-8") as fin:
        lines = fin.readlines()

    all_words = []
    all_coref = []
    current_words = []
    current_coref = []
    for line in lines:
        line = line.strip()
        if not line:
            all_words.append(current_words)
            all_coref.append(current_coref)
            current_words = []
            current_coref = []
            continue
        pieces = line.split("\t")
        current_words.append(pieces[3])
        current_coref.append(pieces[-1])

    if current_words:
        all_words.append(current_words)
        all_coref.append(current_coref)

    return all_words, all_coref

def convert_clusters(filename, corefs):
    sentence_clusters = []
    # current_clusters will be a list of (cluster id, start idx)
    for sent_idx, sentence_coref in enumerate(corefs):
        current_clusters = []
        processed = []
        for word_idx, word_coref in enumerate(sentence_coref):
            new_clusters = []
            if word_coref == '-':
                pieces = []
            else:
                pieces = word_coref.split(";")
            for piece in pieces:
                if not piece.startswith("I-") and not piece.startswith("B-"):
                    raise ValueError("Unexpected coref format %s in document %s" % (word_coref, filename))
                if piece.startswith("B-"):
                    new_clusters.append((int(piece[2:]), word_idx))
                else:
                    assert piece.startswith("I-")
                    cluster_id = int(piece[2:])
                    # this will keep the first cluster found
                    # the effect of this is that when two clusters overlap,
                    # and they happen to be the same cluster id,
                    # they will be nested instead of overlapping past each other
                    for idx, previous_cluster in enumerate(current_clusters):
                        if previous_cluster[0] == cluster_id:
                            break
                    else:
                        raise ValueError("Cluster %s does not continue an existing cluster in %s" % (piece, filename))
                    new_clusters.append(previous_cluster)
                    del current_clusters[idx]

            for cluster, start_idx in current_clusters:
                processed.append((cluster, start_idx, word_idx-1))
            current_clusters = new_clusters
        for cluster, start_idx in current_clusters:
            processed.append((cluster, start_idx, len(sentence_coref)-1))
        # sort by the first word index
        processed = sorted(processed, key=itemgetter(1))
        # TODO: cluster IDs are starting at 1, not 0.
        # that may or may not be relevant
        sentence_clusters.append(processed)
    return sentence_clusters

def main():
    parser = argparse.ArgumentParser(
        prog='Convert Tamil Coref Data',
    )
    parser.add_argument('--no_use_cconj_heads', dest='use_cconj_heads', action='store_false', help="Don't use the conjunction-aware transformation")
    args = parser.parse_args()

    random.seed(1234)

    paths = get_default_paths()
    coref_input_path = paths["COREF_BASE"]
    tamil_base_path = os.path.join(coref_input_path, "tamil", "coref_ta_corrected")
    tamil_glob = os.path.join(tamil_base_path, "*txt")

    filenames = sorted(glob.glob(tamil_glob))
    docs = [read_doc(x) for x in filenames]
    raw_sentences = [doc[0] for doc in docs]
    sentence_clusters = [convert_clusters(filename, doc[1]) for filename, doc in zip(filenames, docs)]

    pipe = stanza.Pipeline("ta", processors="tokenize,pos,lemma,depparse", package="default_accurate", tokenize_pretokenized=True)

    train, dev, test = [], [], []
    for filename, sentences, coref_spans in tqdm(zip(filenames, raw_sentences, sentence_clusters), total=len(filenames)):
        doc_id = filename
        part_id = " "
        sentence_speakers = [[""] * len(sent) for sent in sentences]

        processed = process_document(pipe, doc_id, part_id, sentences, coref_spans, sentence_speakers, use_cconj_heads=args.use_cconj_heads)
        location = random.choices((train, dev, test), weights = (0.8, 0.1, 0.1))[0]
        location.append(processed)

    output_filename = os.path.join(paths["COREF_DATA_DIR"], "ta_kbc.train.json")
    write_json_file(output_filename, train)

    output_filename = os.path.join(paths["COREF_DATA_DIR"], "ta_kbc.dev.json")
    write_json_file(output_filename, dev)

    output_filename = os.path.join(paths["COREF_DATA_DIR"], "ta_kbc.test.json")
    write_json_file(output_filename, test)

if __name__ == '__main__':
    main()



"""
Use the concepts in "Dataset Cartography" and "Mind Your Outliers" to find trees with the least variance over a training run

https://arxiv.org/pdf/2009.10795.pdf
https://arxiv.org/abs/2107.02331

The idea here is that high variance trees are more likely to be wrong in the first place.  Using this will filter a silver dataset to have better trees.

for example:

nlprun -d a6000 -p high "export CLASSPATH=/sailhome/horatio/CoreNLP/classes:/sailhome/horatio/CoreNLP/lib/*:$CLASSPATH; python3 stanza/utils/datasets/constituency/silver_variance.py --eval_file /u/nlp/data/constituency-parser/italian/2024_it_vit_electra/it_silver_0.mrg saved_models/constituency/it_vit.top.each.silver0.constituency_0*0.pt --output_file filtered_silver0.mrg" -o filter.out
"""

import argparse

import logging

import numpy

from stanza.models.common import utils
from stanza.models.common.foundation_cache import FoundationCache
from stanza.models.constituency import retagging
from stanza.models.constituency import tree_reader
from stanza.models.constituency.parser_training import run_dev_set
from stanza.models.constituency.trainer import Trainer
from stanza.models.constituency.utils import retag_trees
from stanza.server.parser_eval import EvaluateParser
from stanza.utils.get_tqdm import get_tqdm

tqdm = get_tqdm()

logger = logging.getLogger('stanza.constituency.trainer')

def parse_args(args=None):
    parser = argparse.ArgumentParser(description="Script to filter trees by how much variance they show over multiple checkpoints of a parser training run.")

    parser.add_argument('--eval_file', type=str, default=None, help='Input file for data loader.')
    parser.add_argument('--output_file', type=str, default=None, help='Output file after sorting by variance.')

    parser.add_argument('--charlm_forward_file', type=str, default=None, help="Exact path to use for forward charlm")
    parser.add_argument('--charlm_backward_file', type=str, default=None, help="Exact path to use for backward charlm")
    parser.add_argument('--wordvec_pretrain_file', type=str, default=None, help='Exact name of the pretrain file to read')

    utils.add_device_args(parser)

    # TODO: use the training scripts to pick the charlm & pretrain if needed
    parser.add_argument('--lang', default='it', help='Language to use')

    parser.add_argument('--eval_batch_size', type=int, default=50, help='How many trees to batch when running eval')
    parser.add_argument('models', type=str, nargs='+', default=None, help="Which model(s) to load")

    parser.add_argument('--keep', type=float, default=0.5, help="How many trees to keep after sorting by variance")
    parser.add_argument('--reverse', default=False, action='store_true', help='Actually, keep the high variance trees')

    retagging.add_retag_args(parser)

    args = vars(parser.parse_args())

    retagging.postprocess_args(args)

    return args

def main():
    args = parse_args()
    retag_pipeline = retagging.build_retag_pipeline(args)
    foundation_cache = retag_pipeline[0].foundation_cache if retag_pipeline else FoundationCache()

    print("Analyzing with the following models:\n  " + "\n  ".join(args['models']))

    treebank = tree_reader.read_treebank(args['eval_file'])
    logger.info("Read %d trees for analysis", len(treebank))

    f1_history = []
    retagged_treebank = None

    chunk_size = 5000
    with EvaluateParser() as evaluator:
        for model_filename in args['models']:
            print("Starting processing with %s" % model_filename)
            trainer = Trainer.load(model_filename, args=args, foundation_cache=foundation_cache)
            if retag_pipeline is not None and retagged_treebank is None:
                retag_method = trainer.model.args['retag_method']
                retag_xpos = trainer.model.args['retag_xpos']
                logger.info("Retagging trees using the %s tags from the %s package...", retag_method, args['retag_package'])
                retagged_treebank = retag_trees(treebank, retag_pipeline, retag_xpos)
                logger.info("Retagging finished")

            current_history = []
            for chunk_start in range(0, len(treebank), chunk_size):
                chunk = treebank[chunk_start:chunk_start+chunk_size]
                retagged_chunk = retagged_treebank[chunk_start:chunk_start+chunk_size] if retagged_treebank else None
                f1, kbestF1, treeF1 = run_dev_set(trainer.model, retagged_chunk, chunk, args, evaluator)
                current_history.extend(treeF1)

            f1_history.append(current_history)

    f1_history = numpy.array(f1_history)
    f1_variance = numpy.var(f1_history, axis=0)
    f1_sorted = sorted([(x, idx) for idx, x in enumerate(f1_variance)], reverse=args['reverse'])

    num_keep = int(len(f1_sorted) * args['keep'])
    with open(args['output_file'], "w", encoding="utf-8") as fout:
        for _, idx in f1_sorted[:num_keep]:
            fout.write(str(treebank[idx]))
            fout.write("\n")

if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod

import logging

import torch
import torch.nn as nn

from stanza.models.common.utils import split_into_batches, sort_with_indices, unsort

"""
A base classifier type

Currently, has the ability to process text or other inputs in a manner
suitable for the particular model type.
In other words, the CNNClassifier processes lists of words,
and the ConstituencyClassifier processes trees
"""

logger = logging.getLogger('stanza')

class BaseClassifier(ABC, nn.Module):
    @abstractmethod
    def extract_sentences(self, doc):
        """
        Extract the sentences or the relevant information in the sentences from a document
        """

    def preprocess_sentences(self, sentences):
        """
        By default, don't do anything
        """
        return sentences

    def label_sentences(self, sentences, batch_size=None):
        """
        Given a list of sentences, return the model's results on that text.
        """
        self.eval()

        sentences = self.preprocess_sentences(sentences)

        if batch_size is None:
            intervals = [(0, len(sentences))]
            orig_idx = None
        else:
            sentences, orig_idx = sort_with_indices(sentences, key=len, reverse=True)
            intervals = split_into_batches(sentences, batch_size)
        labels = []
        for interval in intervals:
            if interval[1] - interval[0] == 0:
                # this can happen for empty text
                continue
            output = self(sentences[interval[0]:interval[1]])
            predicted = torch.argmax(output, dim=1)
            labels.extend(predicted.tolist())

        if orig_idx:
            sentences = unsort(sentences, orig_idx)
            labels = unsort(labels, orig_idx)

        logger.debug("Found labels")
        for (label, sentence) in zip(labels, sentences):
            logger.debug((label, sentence))

        return labels



from enum import Enum

from stanza.models.constituency.dynamic_oracle import advance_past_constituents, find_in_order_constituent_end, find_previous_open, DynamicOracle
from stanza.models.constituency.parse_transitions import Shift, OpenConstituent, CloseConstituent, CompoundUnary, Finalize

def fix_missing_unary_error(gold_transition, pred_transition, gold_sequence, gold_index, root_labels, model, state):
    """
    A CompoundUnary transition was missed after a Shift, but the sequence was continued correctly otherwise
    """
    if not isinstance(gold_transition, CompoundUnary):
        return None

    if pred_transition != gold_sequence[gold_index + 1]:
        return None
    if isinstance(pred_transition, Finalize):
        # this can happen if the entire tree is a single word
        # but it can't be fixed if it means the parser missed the ROOT transition
        return None

    return gold_sequence[:gold_index] + gold_sequence[gold_index+1:]

def fix_wrong_unary_error(gold_transition, pred_transition, gold_sequence, gold_index, root_labels, model, state):
    if not isinstance(gold_transition, CompoundUnary):
        return None

    if not isinstance(pred_transition, CompoundUnary):
        return None

    assert gold_transition != pred_transition

    return gold_sequence[:gold_index] + [pred_transition] + gold_sequence[gold_index+1:]

def fix_spurious_unary_error(gold_transition, pred_transition, gold_sequence, gold_index, root_labels, model, state):
    if isinstance(gold_transition, CompoundUnary):
        return None

    if not isinstance(pred_transition, CompoundUnary):
        return None

    return gold_sequence[:gold_index] + [pred_transition] + gold_sequence[gold_index:]

def fix_open_shift_error(gold_transition, pred_transition, gold_sequence, gold_index, root_labels, model, state):
    """
    Fix a missed Open constituent where we predicted a Shift and the next transition was a Shift

    In fact, the subsequent transition MUST be a Shift with this transition scheme
    """
    if not isinstance(gold_transition, OpenConstituent):
        return None

    if not isinstance(pred_transition, Shift):
        return None

    #if not isinstance(gold_sequence[gold_index+1], Shift):
    #    return None
    assert isinstance(gold_sequence[gold_index+1], Shift)

    # close_index represents the Close for the missing Open
    close_index = advance_past_constituents(gold_sequence, gold_index+1)
    assert close_index is not None
    return gold_sequence[:gold_index] + gold_sequence[gold_index+1:close_index] + gold_sequence[close_index+1:]

def fix_open_open_two_subtrees_error(gold_transition, pred_transition, gold_sequence, gold_index, root_labels, model, state):
    if gold_transition == pred_transition:
        return None

    if not isinstance(gold_transition, OpenConstituent):
        return None
    if not isinstance(pred_transition, OpenConstituent):
        return None

    block_end = find_in_order_constituent_end(gold_sequence, gold_index+1)
    if isinstance(gold_sequence[block_end], Shift):
        # this is a multiple subtrees version of this error
        # we are only skipping the two subtrees errors for now
        return None

    # no fix is possible, so we just return here
    return RepairType.OPEN_OPEN_TWO_SUBTREES_ERROR, None

def fix_open_open_error(gold_transition, pred_transition, gold_sequence, gold_index, root_labels, exactly_three):
    if gold_transition == pred_transition:
        return None

    if not isinstance(gold_transition, OpenConstituent):
        return None
    if not isinstance(pred_transition, OpenConstituent):
        return None

    block_end = find_in_order_constituent_end(gold_sequence, gold_index+1)
    if not isinstance(gold_sequence[block_end], Shift):
        # this is a multiple subtrees version of this error
        # we are only skipping the two subtrees errors for now
        return None

    next_block_end = find_in_order_constituent_end(gold_sequence, block_end+1)
    if exactly_three and isinstance(gold_sequence[next_block_end], Shift):
        # for exactly three subtrees,
        # we can put back the missing open transition
        # and now we have no recall error, only precision error
        # for more than three, we separate that out as an ambiguous choice
        return None
    elif not exactly_three and isinstance(gold_sequence[next_block_end], CloseConstituent):
        # this is ambiguous, but we can still try this fix
        return None

    # at this point, we build a new sequence with the origin constituent inserted
    return gold_sequence[:gold_index] + [pred_transition] + gold_sequence[gold_index+1:block_end] + [CloseConstituent(), gold_transition] + gold_sequence[block_end:]


def fix_open_open_three_subtrees_error(gold_transition, pred_transition, gold_sequence, gold_index, root_labels, model, state):
    return fix_open_open_error(gold_transition, pred_transition, gold_sequence, gold_index, root_labels, exactly_three=True)

def fix_open_open_many_subtrees_error(gold_transition, pred_transition, gold_sequence, gold_index, root_labels, model, state):
    return fix_open_open_error(gold_transition, pred_transition, gold_sequence, gold_index, root_labels, exactly_three=False)

def fix_open_close_error(gold_transition, pred_transition, gold_sequence, gold_index, root_labels, model, state):
    """
    Find the closed bracket, reopen it

    The Open we just missed must be forgotten - it cannot be reopened
    """
    if not isinstance(gold_transition, OpenConstituent):
        return None

    if not isinstance(pred_transition, CloseConstituent):
        return None

    # find the appropriate Open so we can reopen it
    open_idx = find_previous_open(gold_sequence, gold_index)
    # actually, if the Close is legal, this can't happen
    # but it might happen in a unit test which doesn't check legality
    if open_idx is None:
        return None

    # also, since we are punting on the missed Open, we need to skip
    # the Close which would have closed it
    close_idx = advance_past_constituents(gold_sequence, gold_index+1)

    return gold_sequence[:gold_index] + [pred_transition, gold_sequence[open_idx]] + gold_sequence[gold_index+1:close_idx] + gold_sequence[close_idx+1:]

def fix_shift_close_error(gold_transition, pred_transition, gold_sequence, gold_index, root_labels, model, state):
    """
    Find the closed bracket, reopen it
    """
    if not isinstance(gold_transition, Shift):
        return None

    if not isinstance(pred_transition, CloseConstituent):
        return None

    # don't do this at the start or immediately after opening
    if gold_index == 0 or isinstance(gold_sequence[gold_index - 1], OpenConstituent):
        return None

    open_idx = find_previous_open(gold_sequence, gold_index)
    assert open_idx is not None

    return gold_sequence[:gold_index] + [pred_transition, gold_sequence[open_idx]] + gold_sequence[gold_index:]

def fix_shift_open_unambiguous_error(gold_transition, pred_transition, gold_sequence, gold_index, root_labels, model, state):
    if not isinstance(gold_transition, Shift):
        return None

    if not isinstance(pred_transition, OpenConstituent):
        return None

    bracket_end = find_in_order_constituent_end(gold_sequence, gold_index)
    assert bracket_end is not None
    if isinstance(gold_sequence[bracket_end], Shift):
        # this is an ambiguous error
        # multiple possible places to end the wrong constituent
        return None
    assert isinstance(gold_sequence[bracket_end], CloseConstituent)

    return gold_sequence[:gold_index] + [pred_transition] + gold_sequence[gold_index:bracket_end] + [CloseConstituent()] + gold_sequence[bracket_end:]

def fix_close_shift_unambiguous_error(gold_transition, pred_transition, gold_sequence, gold_index, root_labels, model, state):
    if not isinstance(gold_transition, CloseConstituent):
        return None

    if not isinstance(pred_transition, Shift):
        return None
    if not isinstance(gold_sequence[gold_index+1], Shift):
        return None

    bracket_end = find_in_order_constituent_end(gold_sequence, gold_index+1)
    assert bracket_end is not None
    if isinstance(gold_sequence[bracket_end], Shift):
        # this is an ambiguous error
        # multiple possible places to end the wrong constituent
        return None
    assert isinstance(gold_sequence[bracket_end], CloseConstituent)

    return gold_sequence[:gold_index] + gold_sequence[gold_index+1:bracket_end] + [CloseConstituent()] + gold_sequence[bracket_end:]

class RepairType(Enum):
    """
    Keep track of which repair is used, if any, on an incorrect transition

    Effects of different repair types:
      no oracle:                0.9251  0.9226
     +missing_unary:            0.9246  0.9214
     +wrong_unary:              0.9236  0.9213
     +spurious_unary:           0.9247  0.9229
     +open_shift_error:         0.9258  0.9226
     +open_open_two_subtrees:   0.9256  0.9215    # nothing changes with this one...
     +open_open_three_subtrees: 0.9256  0.9226
     +open_open_many_subtrees:  0.9257  0.9234
     +shift_close:              0.9267  0.9250
     +shift_open:               0.9273  0.9247
     +close_shift:              0.9266  0.9229
     +open_close:               0.9267  0.9256
    """
    def __new__(cls, fn, correct=False, debug=False):
        """
        Enumerate values as normal, but also keep a pointer to a function which repairs that kind of error
        """
        value = len(cls.__members__)
        obj = object.__new__(cls)
        obj._value_ = value + 1
        obj.fn = fn
        obj.correct = correct
        obj.debug = debug
        return obj

    @property
    def is_correct(self):
        return self.correct

    # The correct sequence went Shift - Unary - Stuff
    # but the CompoundUnary was missed and Stuff predicted
    # so now we just proceed as if nothing happened
    # note that CompoundUnary happens immediately after a Shift
    # complicated nodes are created with single Open transitions
    MISSING_UNARY_ERROR                    = (fix_missing_unary_error,)

    # Predicted a wrong CompoundUnary.  No way to fix this, so just keep going
    WRONG_UNARY_ERROR                      = (fix_wrong_unary_error,)

    # The correct sequence went Shift - Stuff
    # but instead we predicted a CompoundUnary
    # again, we just keep going
    SPURIOUS_UNARY_ERROR                   = (fix_spurious_unary_error,)

    # Were supposed to open a new constituent,
    # but instead shifted an item onto the stack
    #
    # The missed Open cannot be recovered
    #
    # One could ask, is it possible to open a bigger constituent later,
    # but if the constituent patterns go
    #   X (good open) Y (missed open) Z
    # when we eventually close Y and Z, because of the missed Open,
    # it is guaranteed to capture X as well
    # since it will grab constituents until one left of the previous Open before Y
    #
    # Therefore, in this case, we must simply forget about this Open (recall error)
    OPEN_SHIFT_ERROR                       = (fix_open_shift_error,)

    # With this transition scheme, it is not possible to fix the following pattern:
    #   T1 O_x T2 C -> T1 O_y T2 C
    # seeing as how there are no unary transitions
    # so whatever precision & recall errors are caused by substituting O_x -> O_y
    # (which could include multiple transitions)
    # those errors are unfixable in any way
    OPEN_OPEN_TWO_SUBTREES_ERROR           = (fix_open_open_two_subtrees_error,)

    # With this transition scheme, a three subtree branch with a wrong Open
    # has a non-ambiguous fix
    #   T1 O_x T2 T3 C -> T1 O_y T2 T3 C
    # this can become
    #   T1 O_y T2 C O_x T3 C
    # now there are precision errors from the incorrectly added transition(s),
    # but the correctly replaced transitions are unambiguous
    OPEN_OPEN_THREE_SUBTREES_ERROR         = (fix_open_open_three_subtrees_error,)

    # We were supposed to shift a new item onto the stack,
    # but instead we closed the previous constituent
    # This causes a precision error, but we can avoid the recall error
    # by immediately reopening the closed constituent.
    SHIFT_CLOSE_ERROR                      = (fix_shift_close_error,)

    # We opened a new constituent instead of shifting
    # In the event that the next constituent ends with a close,
    # rather than building another new constituent,
    # then there is no ambiguity
    SHIFT_OPEN_UNAMBIGUOUS_ERROR           = (fix_shift_open_unambiguous_error,)

    # Suppose we were supposed to Close, then Shift
    # but instead we just did a Shift
    # Similar to shift_open_unambiguous, we now have an opened
    # constituent which shouldn't be there
    # We can scroll past the next constituent created to see
    # if the outer constituents close at that point
    # If so, we can close this constituent as well in an unambiguous manner
    # TODO: analyze the case where we were supposed to Close, Open
    # but instead did a Shift
    CLOSE_SHIFT_UNAMBIGUOUS_ERROR          = (fix_close_shift_unambiguous_error,)

    # Supposed to open a new constituent,
    # instead closed an existing constituent
    #
    #  X (good open) Y (open -> close) Z
    #
    # the constituent that should contain Y, Z is unfortunately lost
    # since now the stack has
    #
    #  XY ...
    #
    # furthermore, there is now a precision error for the extra XY
    # constituent that should not exist
    # however, what we can do to minimize further errors is
    # to at least reopen the label between X and Y
    OPEN_CLOSE_ERROR                       = (fix_open_close_error,)

    # this is ambiguous, but we can still try the same fix as three_subtrees (see above)
    OPEN_OPEN_MANY_SUBTREES_ERROR          = (fix_open_open_many_subtrees_error,)

    CORRECT                                = (None, True)

    UNKNOWN                                = None


class InOrderCompoundOracle(DynamicOracle):
    def __init__(self, root_labels, oracle_level, additional_oracle_levels, deactivated_oracle_levels):
        super().__init__(root_labels, oracle_level, RepairType, additional_oracle_levels, deactivated_oracle_levels)



"""
Test a couple different classes of trees to check the output of the VIT conversion

A couple representative trees are included, but hopefully not enough
to be a problem in terms of our license.

One of the tests is currently disabled as it relies on tregex & tsurgeon features
not yet released
"""

import io
import os
import tempfile

import pytest

from stanza.server import tsurgeon
from stanza.utils.conll import CoNLL
from stanza.utils.datasets.constituency import convert_it_vit

pytestmark = [pytest.mark.pipeline, pytest.mark.travis]

# just a sample!  don't sue us please
CON_SAMPLE = """
#ID=sent_00002	cp-[sp-[part-negli, sn-[sa-[ag-ultimi], nt-anni]], f-[sn-[art-la, n-dinamica, spd-[partd-dei, sn-[n-polo_di_attrazione]]], ibar-[ause-è, ausep-stata, savv-[savv-[avv-sempre], avv-più], vppt-caratterizzata], compin-[spda-[partda-dall, sn-[n-emergere, spd-[pd-di, sn-[art-una, sa-[ag-crescente], n-concorrenza, f2-[rel-che, f-[ibar-[clit-si, ause-è, avv-progressivamente, vppin-spostata], compin-[spda-[partda-dalle, sn-[sa-[ag-singole], n-imprese]], sp-[part-ai, sn-[n-sistemi, sa-[coord-[ag-economici, cong-e, ag-territoriali]]]], fp-[punt-',', sv5-[vgt-determinando, compt-[sn-[art-l_, nf-esigenza, spd-[pd-di, sn-[art-una, n-riconsiderazione, spd-[partd-dei, sn-[n-rapporti, sv3-[ppre-esistenti, compin-[sp-[p-tra, sn-[n-soggetti, sa-[ag-produttivi]]], cong-e, sn-[n-ambiente, f2-[sp-[p-in, sn-[relob-cui]], f-[sn-[deit-questi], ibar-[vin-operano, punto-.]]]]]]]]]]]]]]]]]]]]]]]]

#ID=sent_00318	dirsp-[fc-[congf-tuttavia, f-[sn-[sq-[ind-qualche], n-problema], ir_infl-[vsupir-potrebbe, vcl-esserci], compc-[clit-ci, sp-[p-per, sn-[art-la, n-commissione, sa-[ag-esteri], f2-[sp-[part-alla, relob-cui, sn-[n-presidenza]], f-[ibar-[vc-è], compc-[sn-[n-candidato], sn-[art-l, n-esponente, spd-[pd-di, sn-[mw-Alleanza, npro-Nazionale]], sn-[mw-Mirko, nh-Tremaglia]]]]]]]]]], dirs-':', f3-[sn-[art-una, n-candidatura, sc-[q-più, sa-[ppas-subìta], sc-[ccong-che, sa-[ppas-gradita]], compt-[spda-[partda-dalla, sn-[mw-Lega, npro-Nord, punt-',', f2-[rel-che, fc-[congf-tuttavia, f-[ir_infl-[vsupir-dovrebbe, vit-rispettare], compt-[sn-[art-gli, n-accordi]]]]]]]]]], punto-.]]

#ID=sent_00589	f-[sn-[art-l, n-ottimismo, spd-[pd-di, sn-[nh-Kantor]]], ir_infl-[vsupir-potrebbe, congf-però, vcl-rivelarsi], compc-[sn-[in-ancora, art-una, nt-volta], sa-[ag-prematuro]], punto-.]
"""

UD_SAMPLE = """
# sent_id = VIT-2
# text = Negli ultimi anni la dinamica dei polo di attrazione è stata sempre più caratterizzata dall'emergere di una crescente concorrenza che si è progressivamente spostata dalle singole imprese ai sistemi economici e territoriali, determinando l'esigenza di una riconsiderazione dei rapporti esistenti tra soggetti produttivi e ambiente in cui questi operano.
1-2	Negli	_	_	_	_	_	_	_	_
1	In	in	ADP	E	_	4	case	_	_
2	gli	il	DET	RD	Definite=Def|Gender=Masc|Number=Plur|PronType=Art	4	det	_	_
3	ultimi	ultimo	ADJ	A	Gender=Masc|Number=Plur	4	amod	_	_
4	anni	anno	NOUN	S	Gender=Masc|Number=Plur	16	obl	_	_
5	la	il	DET	RD	Definite=Def|Gender=Fem|Number=Sing|PronType=Art	6	det	_	_
6	dinamica	dinamica	NOUN	S	Gender=Fem|Number=Sing	16	nsubj:pass	_	_
7-8	dei	_	_	_	_	_	_	_	_
7	di	di	ADP	E	_	9	case	_	_
8	i	il	DET	RD	Definite=Def|Gender=Masc|Number=Plur|PronType=Art	9	det	_	_
9	polo	polo	NOUN	S	Gender=Masc|Number=Sing	6	nmod	_	_
10	di	di	ADP	E	_	11	case	_	_
11	attrazione	attrazione	NOUN	S	Gender=Fem|Number=Sing	9	nmod	_	_
12	è	essere	AUX	VA	Mood=Ind|Number=Sing|Person=3|Tense=Pres|VerbForm=Fin	16	aux	_	_
13	stata	essere	AUX	VA	Gender=Fem|Number=Sing|Tense=Past|VerbForm=Part	16	aux:pass	_	_
14	sempre	sempre	ADV	B	_	15	advmod	_	_
15	più	più	ADV	B	_	16	advmod	_	_
16	caratterizzata	caratterizzare	VERB	V	Gender=Fem|Number=Sing|Tense=Past|VerbForm=Part	0	root	_	_
17-18	dall'	_	_	_	_	_	_	_	SpaceAfter=No
17	da	da	ADP	E	_	19	case	_	_
18	l'	il	DET	RD	Definite=Def|Number=Sing|PronType=Art	19	det	_	_
19	emergere	emergere	NOUN	S	Gender=Masc|Number=Sing	16	obl	_	_
20	di	di	ADP	E	_	23	case	_	_
21	una	uno	DET	RI	Definite=Ind|Gender=Fem|Number=Sing|PronType=Art	23	det	_	_
22	crescente	crescente	ADJ	A	Number=Sing	23	amod	_	_
23	concorrenza	concorrenza	NOUN	S	Gender=Fem|Number=Sing	19	nmod	_	_
24	che	che	PRON	PR	PronType=Rel	28	nsubj	_	_
25	si	si	PRON	PC	Clitic=Yes|Person=3|PronType=Prs	28	expl	_	_
26	è	essere	AUX	VA	Mood=Ind|Number=Sing|Person=3|Tense=Pres|VerbForm=Fin	28	aux	_	_
27	progressivamente	progressivamente	ADV	B	_	28	advmod	_	_
28	spostata	spostare	VERB	V	Gender=Fem|Number=Sing|Tense=Past|VerbForm=Part	23	acl:relcl	_	_
29-30	dalle	_	_	_	_	_	_	_	_
29	da	da	ADP	E	_	32	case	_	_
30	le	il	DET	RD	Definite=Def|Gender=Fem|Number=Plur|PronType=Art	32	det	_	_
31	singole	singolo	ADJ	A	Gender=Fem|Number=Plur	32	amod	_	_
32	imprese	impresa	NOUN	S	Gender=Fem|Number=Plur	28	obl	_	_
33-34	ai	_	_	_	_	_	_	_	_
33	a	a	ADP	E	_	35	case	_	_
34	i	il	DET	RD	Definite=Def|Gender=Masc|Number=Plur|PronType=Art	35	det	_	_
35	sistemi	sistema	NOUN	S	Gender=Masc|Number=Plur	28	obl	_	_
36	economici	economico	ADJ	A	Gender=Masc|Number=Plur	35	amod	_	_
37	e	e	CCONJ	CC	_	38	cc	_	_
38	territoriali	territoriale	ADJ	A	Number=Plur	36	conj	_	SpaceAfter=No
39	,	,	PUNCT	FF	_	28	punct	_	_
40	determinando	determinare	VERB	V	VerbForm=Ger	28	advcl	_	_
41	l'	il	DET	RD	Definite=Def|Number=Sing|PronType=Art	42	det	_	SpaceAfter=No
42	esigenza	esigenza	NOUN	S	Gender=Fem|Number=Sing	40	obj	_	_
43	di	di	ADP	E	_	45	case	_	_
44	una	uno	DET	RI	Definite=Ind|Gender=Fem|Number=Sing|PronType=Art	45	det	_	_
45	riconsiderazione	riconsiderazione	NOUN	S	Gender=Fem|Number=Sing	42	nmod	_	_
46-47	dei	_	_	_	_	_	_	_	_
46	di	di	ADP	E	_	48	case	_	_
47	i	il	DET	RD	Definite=Def|Gender=Masc|Number=Plur|PronType=Art	48	det	_	_
48	rapporti	rapporto	NOUN	S	Gender=Masc|Number=Plur	45	nmod	_	_
49	esistenti	esistente	VERB	V	Number=Plur	48	acl	_	_
50	tra	tra	ADP	E	_	51	case	_	_
51	soggetti	soggetto	NOUN	S	Gender=Masc|Number=Plur	49	obl	_	_
52	produttivi	produttivo	ADJ	A	Gender=Masc|Number=Plur	51	amod	_	_
53	e	e	CCONJ	CC	_	54	cc	_	_
54	ambiente	ambiente	NOUN	S	Gender=Masc|Number=Sing	51	conj	_	_
55	in	in	ADP	E	_	56	case	_	_
56	cui	cui	PRON	PR	PronType=Rel	58	obl	_	_
57	questi	questo	PRON	PD	Gender=Masc|Number=Plur|PronType=Dem	58	nsubj	_	_
58	operano	operare	VERB	V	Mood=Ind|Number=Plur|Person=3|Tense=Pres|VerbForm=Fin	54	acl:relcl	_	SpaceAfter=No
59	.	.	PUNCT	FS	_	16	punct	_	_

# sent_id = VIT-318
# text = Tuttavia qualche problema potrebbe esserci per la commissione esteri alla cui presidenza è candidato l'esponente di Alleanza Nazionale Mirko Tremaglia: una candidatura più subìta che gradita dalla Lega Nord, che tuttavia dovrebbe rispettare gli accordi.
1	Tuttavia	tuttavia	CCONJ	CC	_	5	cc	_	_
2	qualche	qualche	DET	DI	Number=Sing|PronType=Ind	3	det	_	_
3	problema	problema	NOUN	S	Gender=Masc|Number=Sing	5	nsubj	_	_
4	potrebbe	potere	AUX	VA	Mood=Cnd|Number=Sing|Person=3|Tense=Pres|VerbForm=Fin	5	aux	_	_
5-6	esserci	_	_	_	_	_	_	_	_
5	esser	essere	VERB	V	VerbForm=Inf	0	root	_	_
6	ci	ci	PRON	PC	Clitic=Yes|Number=Plur|Person=1|PronType=Prs	5	expl	_	_
7	per	per	ADP	E	_	9	case	_	_
8	la	il	DET	RD	Definite=Def|Gender=Fem|Number=Sing|PronType=Art	9	det	_	_
9	commissione	commissione	NOUN	S	Gender=Fem|Number=Sing	5	obl	_	_
10	esteri	estero	ADJ	A	Gender=Masc|Number=Plur	9	amod	_	_
11-12	alla	_	_	_	_	_	_	_	_
11	a	a	ADP	E	_	14	case	_	_
12	la	il	DET	RD	Definite=Def|Gender=Fem|Number=Sing|PronType=Art	14	det	_	_
13	cui	cui	DET	DR	PronType=Rel	14	det:poss	_	_
14	presidenza	presidenza	NOUN	S	Gender=Fem|Number=Sing	16	obl	_	_
15	è	essere	AUX	VA	Mood=Ind|Number=Sing|Person=3|Tense=Pres|VerbForm=Fin	16	aux:pass	_	_
16	candidato	candidare	VERB	V	Gender=Masc|Number=Sing|Tense=Past|VerbForm=Part	9	acl:relcl	_	_
17	l'	il	DET	RD	Definite=Def|Number=Sing|PronType=Art	18	det	_	SpaceAfter=No
18	esponente	esponente	NOUN	S	Number=Sing	16	nsubj:pass	_	_
19	di	di	ADP	E	_	20	case	_	_
20	Alleanza	Alleanza	PROPN	SP	_	18	nmod	_	_
21	Nazionale	Nazionale	PROPN	SP	_	20	flat:name	_	_
22	Mirko	Mirko	PROPN	SP	_	18	nmod	_	_
23	Tremaglia	Tremaglia	PROPN	SP	_	22	flat:name	_	SpaceAfter=No
24	:	:	PUNCT	FC	_	22	punct	_	_
25	una	uno	DET	RI	Definite=Ind|Gender=Fem|Number=Sing|PronType=Art	26	det	_	_
26	candidatura	candidatura	NOUN	S	Gender=Fem|Number=Sing	22	appos	_	_
27	più	più	ADV	B	_	28	advmod	_	_
28	subìta	subire	VERB	V	Gender=Fem|Number=Sing|Tense=Past|VerbForm=Part	26	advcl	_	_
29	che	che	CCONJ	CC	_	30	cc	_	_
30	gradita	gradito	ADJ	A	Gender=Fem|Number=Sing	28	amod	_	_
31-32	dalla	_	_	_	_	_	_	_	_
31	da	da	ADP	E	_	33	case	_	_
32	la	il	DET	RD	Definite=Def|Gender=Fem|Number=Sing|PronType=Art	33	det	_	_
33	Lega	Lega	PROPN	SP	_	28	obl:agent	_	_
34	Nord	Nord	PROPN	SP	_	33	flat:name	_	SpaceAfter=No
35	,	,	PUNCT	FC	_	33	punct	_	_
36	che	che	PRON	PR	PronType=Rel	39	nsubj	_	_
37	tuttavia	tuttavia	CCONJ	CC	_	39	cc	_	_
38	dovrebbe	dovere	AUX	VM	Mood=Cnd|Number=Sing|Person=3|Tense=Pres|VerbForm=Fin	39	aux	_	_
39	rispettare	rispettare	VERB	V	VerbForm=Inf	33	acl:relcl	_	_
40	gli	il	DET	RD	Definite=Def|Gender=Masc|Number=Plur|PronType=Art	41	det	_	_
41	accordi	accordio	NOUN	S	Gender=Masc|Number=Plur	39	obj	_	SpaceAfter=No
42	.	.	PUNCT	FS	_	5	punct	_	_

# sent_id = VIT-591
# text = L'ottimismo di Kantor potrebbe però rivelarsi ancora una volta prematuro.
1	L'	il	DET	RD	Definite=Def|Number=Sing|PronType=Art	2	det	_	SpaceAfter=No
2	ottimismo	ottimismo	NOUN	S	Gender=Masc|Number=Sing	7	nsubj	_	_
3	di	di	ADP	E	_	4	case	_	_
4	Kantor	Kantor	PROPN	SP	_	2	nmod	_	_
5	potrebbe	potere	AUX	VM	Mood=Cnd|Number=Sing|Person=3|Tense=Pres|VerbForm=Fin	7	aux	_	_
6	però	però	ADV	B	_	7	advmod	_	_
7-8	rivelarsi	_	_	_	_	_	_	_	_
7	rivelar	rivelare	VERB	V	VerbForm=Inf	0	root	_	_
8	si	si	PRON	PC	Clitic=Yes|Person=3|PronType=Prs	7	expl	_	_
9	ancora	ancora	ADV	B	_	7	advmod	_	_
10	una	uno	DET	RI	Definite=Ind|Gender=Fem|Number=Sing|PronType=Art	11	det	_	_
11	volta	volta	NOUN	S	Gender=Fem|Number=Sing	7	obl	_	_
12	prematuro	prematuro	ADJ	A	Gender=Masc|Number=Sing	7	xcomp	_	SpaceAfter=No
13	.	.	PUNCT	FS	_	7	punct	_	_
"""


def test_process_mwts():
    # dei appears multiple times
    # the verb/pron esserci will be ignored
    expected_mwts = {'Negli': ('In', 'gli'), 'dei': ('di', 'i'), "dall'": ('da', "l'"), 'dalle': ('da', 'le'), 'ai': ('a', 'i'), 'alla': ('a', 'la'), 'dalla': ('da', 'la')}

    ud_train_data = CoNLL.conll2doc(input_str=UD_SAMPLE)

    mwts = convert_it_vit.get_mwt(ud_train_data)
    assert expected_mwts == mwts

def test_raw_tree():
    con_sentences = convert_it_vit.read_constituency_sentences(io.StringIO(CON_SAMPLE))
    expected_ids = ["#ID=sent_00002", "#ID=sent_00318", "#ID=sent_00589"]
    expected_trees = ["(ROOT (cp (sp (part negli) (sn (sa (ag ultimi)) (nt anni))) (f (sn (art la) (n dinamica) (spd (partd dei) (sn (n polo) (n di) (n attrazione)))) (ibar (ause è) (ausep stata) (savv (savv (avv sempre)) (avv più)) (vppt caratterizzata)) (compin (spda (partda dall) (sn (n emergere) (spd (pd di) (sn (art una) (sa (ag crescente)) (n concorrenza) (f2 (rel che) (f (ibar (clit si) (ause è) (avv progressivamente) (vppin spostata)) (compin (spda (partda dalle) (sn (sa (ag singole)) (n imprese))) (sp (part ai) (sn (n sistemi) (sa (coord (ag economici) (cong e) (ag territoriali))))) (fp (punt ,) (sv5 (vgt determinando) (compt (sn (art l') (nf esigenza) (spd (pd di) (sn (art una) (n riconsiderazione) (spd (partd dei) (sn (n rapporti) (sv3 (ppre esistenti) (compin (sp (p tra) (sn (n soggetti) (sa (ag produttivi)))) (cong e) (sn (n ambiente) (f2 (sp (p in) (sn (relob cui))) (f (sn (deit questi)) (ibar (vin operano) (punto .))))))))))))))))))))))))))",
                      "(ROOT (dirsp (fc (congf tuttavia) (f (sn (sq (ind qualche)) (n problema)) (ir_infl (vsupir potrebbe) (vcl esserci)) (compc (clit ci) (sp (p per) (sn (art la) (n commissione) (sa (ag esteri)) (f2 (sp (part alla) (relob cui) (sn (n presidenza))) (f (ibar (vc è)) (compc (sn (n candidato)) (sn (art l) (n esponente) (spd (pd di) (sn (mw Alleanza) (npro Nazionale))) (sn (mw Mirko) (nh Tremaglia))))))))))) (dirs :) (f3 (sn (art una) (n candidatura) (sc (q più) (sa (ppas subìta)) (sc (ccong che) (sa (ppas gradita))) (compt (spda (partda dalla) (sn (mw Lega) (npro Nord) (punt ,) (f2 (rel che) (fc (congf tuttavia) (f (ir_infl (vsupir dovrebbe) (vit rispettare)) (compt (sn (art gli) (n accordi))))))))))) (punto .))))",
                      "(ROOT (f (sn (art l) (n ottimismo) (spd (pd di) (sn (nh Kantor)))) (ir_infl (vsupir potrebbe) (congf però) (vcl rivelarsi)) (compc (sn (in ancora) (art una) (nt volta)) (sa (ag prematuro))) (punto .)))"]
    assert len(con_sentences) == 3
    for sentence, expected_id, expected_tree in zip(con_sentences, expected_ids, expected_trees):
        assert sentence[0] == expected_id
        tree = convert_it_vit.raw_tree(sentence[1])
        assert str(tree) == expected_tree

def test_update_mwts():
    con_sentences = convert_it_vit.read_constituency_sentences(io.StringIO(CON_SAMPLE))
    ud_train_data = CoNLL.conll2doc(input_str=UD_SAMPLE)
    mwt_map = convert_it_vit.get_mwt(ud_train_data)
    expected_trees=["(ROOT (cp (sp (part In) (sn (art gli) (sa (ag ultimi)) (nt anni))) (f (sn (art la) (n dinamica) (spd (partd di) (sn (art i) (n polo) (n di) (n attrazione)))) (ibar (ause è) (ausep stata) (savv (savv (avv sempre)) (avv più)) (vppt caratterizzata)) (compin (spda (partda da) (sn (art l') (n emergere) (spd (pd di) (sn (art una) (sa (ag crescente)) (n concorrenza) (f2 (rel che) (f (ibar (clit si) (ause è) (avv progressivamente) (vppin spostata)) (compin (spda (partda da) (sn (art le) (sa (ag singole)) (n imprese))) (sp (part a) (sn (art i) (n sistemi) (sa (coord (ag economici) (cong e) (ag territoriali))))) (fp (punt ,) (sv5 (vgt determinando) (compt (sn (art l') (nf esigenza) (spd (pd di) (sn (art una) (n riconsiderazione) (spd (partd di) (sn (art i) (n rapporti) (sv3 (ppre esistenti) (compin (sp (p tra) (sn (n soggetti) (sa (ag produttivi)))) (cong e) (sn (n ambiente) (f2 (sp (p in) (sn (relob cui))) (f (sn (deit questi)) (ibar (vin operano) (punto .))))))))))))))))))))))))))",
                    "(ROOT (dirsp (fc (congf tuttavia) (f (sn (sq (ind qualche)) (n problema)) (ir_infl (vsupir potrebbe) (vcl esserci)) (compc (clit ci) (sp (p per) (sn (art la) (n commissione) (sa (ag esteri)) (f2 (sp (part a) (art la) (relob cui) (sn (n presidenza))) (f (ibar (vc è)) (compc (sn (n candidato)) (sn (art l) (n esponente) (spd (pd di) (sn (mw Alleanza) (npro Nazionale))) (sn (mw Mirko) (nh Tremaglia))))))))))) (dirs :) (f3 (sn (art una) (n candidatura) (sc (q più) (sa (ppas subìta)) (sc (ccong che) (sa (ppas gradita))) (compt (spda (partda da) (sn (art la) (mw Lega) (npro Nord) (punt ,) (f2 (rel che) (fc (congf tuttavia) (f (ir_infl (vsupir dovrebbe) (vit rispettare)) (compt (sn (art gli) (n accordi))))))))))) (punto .))))",
                    "(ROOT (f (sn (art l) (n ottimismo) (spd (pd di) (sn (nh Kantor)))) (ir_infl (vsupir potrebbe) (congf però) (vcl rivelarsi)) (compc (clit si) (sn (in ancora) (art una) (nt volta)) (sa (ag prematuro))) (punto .)))"]
    with tsurgeon.Tsurgeon() as tsurgeon_processor:
        for con_sentence, ud_sentence, expected_tree in zip(con_sentences, ud_train_data.sentences, expected_trees):
            con_tree = convert_it_vit.raw_tree(con_sentence[1])
            updated_tree, _ = convert_it_vit.update_mwts_and_special_cases(con_tree, ud_sentence, mwt_map, tsurgeon_processor)
            assert str(updated_tree) == expected_tree


CON_PERCENT_SAMPLE = """
ID#sent_00020 f-[sn-[art-il, n-tesoro], ibar-[vt-mette], compt-[sp-[part-sul, sn-[n-mercato]], sn-[art-il, num-51%, sp-[p-a, sn-[num-2, n-lire]], sp-[p-per, sn-[n-azione]]]], punto-.]
ID#sent_00022 dirsp-[f3-[sn-[art-le, n-novità]], dirs-':', f3-[coord-[sn-[n-voto, spd-[pd-di, sn-[n-lista]]], cong-e, sn-[n-tetto, sp-[part-agli, sn-[n-acquisti]], sv3-[vppt-limitato, comppas-[sp-[part-allo, sn-[num-0/5%]]]]]], punto-.]]
ID#sent_00517 dirsp-[fc-[f-[sn-[art-l, n-aumento, sa-[ag-mensile], spd-[pd-di, sn-[nt-aprile]]], ibar-[ause-è, vppc-stato], compc-[sq-[q-dell_, sn-[num-1/3%]], sp-[p-contro, sn-[art-lo, num-0/7/0/8%, spd-[partd-degli, sn-[sa-[ag-ultimi], num-due, sn-[nt-mesi]]]]]]]]]
ID#sent_01117 fc-[f-[sn-[art-La, sa-[ag-crescente], n-ripresa, spd-[partd-dei, sn-[n-beni, spd-[pd-di, sn-[n-consumo]]]]], ibar-[vin-deriva], savv-[avv-esclusivamente], compin-[spda-[partda-dal, sn-[n-miglioramento, f2-[spd-[pd-di, sn-[relob-cui]], f-[ibar-[ausa-hanno, vppin-beneficiato], compin-[sn-[n-beni, coord-[sa-[ag-durevoli, fp-[par-'(', sn-[num-plus4/5%], par-')']], cong-e, sa-[ag-semidurevoli, fp-[par-'(', sn-[num-plus1/5%], par-')']]]]]]]]]]], punt-',', fs-[cosu-mentre, f-[sn-[art-i, n-beni, sa-[neg-non, ag-durevoli], fp-[par-'(', sn-[num-min1%], par-')']], ibar-[vt-accusano], cong-ancora, compt-[sn-[art-un, sa-[ag-evidente], n-ritardo]]]], punto-.]
"""

CON_PERCENT_LEAVES = [
    ['il', 'tesoro', 'mette', 'sul', 'mercato', 'il', '51', '%%', 'a', '2', 'lire', 'per', 'azione', '.'],
    ['le', 'novità', ':', 'voto', 'di', 'lista', 'e', 'tetto', 'agli', 'acquisti', 'limitato', 'allo', '0,5', '%%', '.'],
    ['l', 'aumento', 'mensile', 'di', 'aprile', 'è', 'stato', "dell'", '1,3', '%%', 'contro', 'lo', '0/7,0/8', '%%', 'degli', 'ultimi', 'due', 'mesi'],
    # the plus and min look bad, but they get cleaned up when merging with the UD version of the dataset
    ['La', 'crescente', 'ripresa', 'dei', 'beni', 'di', 'consumo', 'deriva', 'esclusivamente', 'dal', 'miglioramento', 'di', 'cui', 'hanno', 'beneficiato', 'beni', 'durevoli', '(', 'plus4,5', '%%', ')', 'e', 'semidurevoli', '(', 'plus1,5', '%%', ')', ',', 'mentre', 'i', 'beni', 'non', 'durevoli', '(', 'min1', '%%', ')', 'accusano', 'ancora', 'un', 'evidente', 'ritardo', '.']
]

def test_read_percent():
    con_sentences = convert_it_vit.read_constituency_sentences(io.StringIO(CON_PERCENT_SAMPLE))
    assert len(con_sentences) == len(CON_PERCENT_LEAVES)
    for (_, raw_tree), expected_leaves in zip(con_sentences, CON_PERCENT_LEAVES):
        tree = convert_it_vit.raw_tree(raw_tree)
        words = tree.leaf_labels()
        if expected_leaves is None:
            print(words)
        else:
            assert words == expected_leaves



import logging
import os

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pad_packed_sequence, pack_padded_sequence, pack_sequence, pad_sequence, PackedSequence

from stanza.models.common.bert_embedding import extract_bert_embeddings
from stanza.models.common.biaffine import BiaffineScorer
from stanza.models.common.foundation_cache import load_bert, load_charlm
from stanza.models.common.hlstm import HighwayLSTM
from stanza.models.common.dropout import WordDropout
from stanza.models.common.utils import attach_bert_model
from stanza.models.common.vocab import CompositeVocab
from stanza.models.common.char_model import CharacterModel
from stanza.models.common import utils

logger = logging.getLogger('stanza')

class Tagger(nn.Module):
    def __init__(self, args, vocab, emb_matrix=None, share_hid=False, foundation_cache=None, bert_model=None, bert_tokenizer=None, force_bert_saved=False, peft_name=None):
        super().__init__()

        self.vocab = vocab
        self.args = args
        self.share_hid = share_hid
        self.unsaved_modules = []

        # input layers
        input_size = 0
        if self.args['word_emb_dim'] > 0:
            # frequent word embeddings
            self.word_emb = nn.Embedding(len(vocab['word']), self.args['word_emb_dim'], padding_idx=0)
            input_size += self.args['word_emb_dim']

        if not share_hid:
            # upos embeddings
            self.upos_emb = nn.Embedding(len(vocab['upos']), self.args['tag_emb_dim'], padding_idx=0)

        if self.args['char'] and self.args['char_emb_dim'] > 0:
            if self.args.get('charlm', None):
                if args['charlm_forward_file'] is None or not os.path.exists(args['charlm_forward_file']):
                    raise FileNotFoundError('Could not find forward character model: {}  Please specify with --charlm_forward_file'.format(args['charlm_forward_file']))
                if args['charlm_backward_file'] is None or not os.path.exists(args['charlm_backward_file']):
                    raise FileNotFoundError('Could not find backward character model: {}  Please specify with --charlm_backward_file'.format(args['charlm_backward_file']))
                logger.debug("POS model loading charmodels: %s and %s", args['charlm_forward_file'], args['charlm_backward_file'])
                self.add_unsaved_module('charmodel_forward', load_charlm(args['charlm_forward_file'], foundation_cache=foundation_cache))
                self.add_unsaved_module('charmodel_backward', load_charlm(args['charlm_backward_file'], foundation_cache=foundation_cache))
                # optionally add a input transformation layer
                if self.args.get('charlm_transform_dim', 0):
                    self.charmodel_forward_transform = nn.Linear(self.charmodel_forward.hidden_dim(), self.args['charlm_transform_dim'], bias=False)
                    self.charmodel_backward_transform = nn.Linear(self.charmodel_backward.hidden_dim(), self.args['charlm_transform_dim'], bias=False)
                    input_size += self.args['charlm_transform_dim'] * 2
                else:
                    self.charmodel_forward_transform = None
                    self.charmodel_backward_transform = None
                    input_size += self.charmodel_forward.hidden_dim() + self.charmodel_backward.hidden_dim()
            else:
                bidirectional = args.get('char_bidirectional', False)
                self.charmodel = CharacterModel(args, vocab, bidirectional=bidirectional)
                if bidirectional:
                    self.trans_char = nn.Linear(self.args['char_hidden_dim'] * 2, self.args['transformed_dim'], bias=False)
                else:
                    self.trans_char = nn.Linear(self.args['char_hidden_dim'], self.args['transformed_dim'], bias=False)
                input_size += self.args['transformed_dim']

        self.peft_name = peft_name
        attach_bert_model(self, bert_model, bert_tokenizer, self.args.get('use_peft', False), force_bert_saved)
        if self.args.get('bert_model', None):
            # TODO: refactor bert_hidden_layers between the different models
            if args.get('bert_hidden_layers', False):
                # The average will be offset by 1/N so that the default zeros
                # represents an average of the N layers
                self.bert_layer_mix = nn.Linear(args['bert_hidden_layers'], 1, bias=False)
                nn.init.zeros_(self.bert_layer_mix.weight)
            else:
                # an average of layers 2, 3, 4 will be used
                # (for historic reasons)
                self.bert_layer_mix = None
            input_size += self.bert_model.config.hidden_size

        if self.args['pretrain']:
            # pretrained embeddings, by default this won't be saved into model file
            self.add_unsaved_module('pretrained_emb', nn.Embedding.from_pretrained(emb_matrix, freeze=True))
            self.trans_pretrained = nn.Linear(emb_matrix.shape[1], self.args['transformed_dim'], bias=False)
            input_size += self.args['transformed_dim']
        
        # recurrent layers
        self.taggerlstm = HighwayLSTM(input_size, self.args['hidden_dim'], self.args['num_layers'], batch_first=True, bidirectional=True, dropout=self.args['dropout'], rec_dropout=self.args['rec_dropout'], highway_func=torch.tanh)
        self.drop_replacement = nn.Parameter(torch.randn(input_size) / np.sqrt(input_size))
        self.taggerlstm_h_init = nn.Parameter(torch.zeros(2 * self.args['num_layers'], 1, self.args['hidden_dim']))
        self.taggerlstm_c_init = nn.Parameter(torch.zeros(2 * self.args['num_layers'], 1, self.args['hidden_dim']))

        # classifiers
        self.upos_hid = nn.Linear(self.args['hidden_dim'] * 2, self.args['deep_biaff_hidden_dim'])
        self.upos_clf = nn.Linear(self.args['deep_biaff_hidden_dim'], len(vocab['upos']))
        self.upos_clf.weight.data.zero_()
        self.upos_clf.bias.data.zero_()

        if share_hid:
            clf_constructor = lambda insize, outsize: nn.Linear(insize, outsize)
        else:
            self.xpos_hid = nn.Linear(self.args['hidden_dim'] * 2, self.args['deep_biaff_hidden_dim'] if not isinstance(vocab['xpos'], CompositeVocab) else self.args['composite_deep_biaff_hidden_dim'])
            self.ufeats_hid = nn.Linear(self.args['hidden_dim'] * 2, self.args['composite_deep_biaff_hidden_dim'])
            clf_constructor = lambda insize, outsize: BiaffineScorer(insize, self.args['tag_emb_dim'], outsize)

        if isinstance(vocab['xpos'], CompositeVocab):
            self.xpos_clf = nn.ModuleList()
            for l in vocab['xpos'].lens():
                self.xpos_clf.append(clf_constructor(self.args['composite_deep_biaff_hidden_dim'], l))
        else:
            self.xpos_clf = clf_constructor(self.args['deep_biaff_hidden_dim'], len(vocab['xpos']))
            if share_hid:
                self.xpos_clf.weight.data.zero_()
                self.xpos_clf.bias.data.zero_()

        self.ufeats_clf = nn.ModuleList()
        for l in vocab['feats'].lens():
            if share_hid:
                self.ufeats_clf.append(clf_constructor(self.args['deep_biaff_hidden_dim'], l))
                self.ufeats_clf[-1].weight.data.zero_()
                self.ufeats_clf[-1].bias.data.zero_()
            else:
                self.ufeats_clf.append(clf_constructor(self.args['composite_deep_biaff_hidden_dim'], l))

        # criterion
        self.crit = nn.CrossEntropyLoss(ignore_index=0) # ignore padding

        self.drop = nn.Dropout(args['dropout'])
        self.worddrop = WordDropout(args['word_dropout'])

    def add_unsaved_module(self, name, module):
        self.unsaved_modules += [name]
        setattr(self, name, module)

    def log_norms(self):
        utils.log_norms(self)

    def forward(self, word, word_mask, wordchars, wordchars_mask, upos, xpos, ufeats, pretrained, word_orig_idx, sentlens, wordlens, text):
        
        def pack(x):
            return pack_padded_sequence(x, sentlens, batch_first=True)

        inputs = []
        if self.args['word_emb_dim'] > 0:
            word_emb = self.word_emb(word)
            word_emb = pack(word_emb)
            inputs += [word_emb]

        if self.args['pretrain']:
            pretrained_emb = self.pretrained_emb(pretrained)
            pretrained_emb = self.trans_pretrained(pretrained_emb)
            pretrained_emb = pack(pretrained_emb)
            inputs += [pretrained_emb]

        def pad(x):
            return pad_packed_sequence(PackedSequence(x, inputs[0].batch_sizes), batch_first=True)[0]

        if self.args['char'] and self.args['char_emb_dim'] > 0:
            if self.args.get('charlm', None):
                all_forward_chars = self.charmodel_forward.build_char_representation(text)
                assert isinstance(all_forward_chars, list)
                if self.charmodel_forward_transform is not None:
                    all_forward_chars = [self.charmodel_forward_transform(x) for x in all_forward_chars]
                all_forward_chars = pack(pad_sequence(all_forward_chars, batch_first=True))

                all_backward_chars = self.charmodel_backward.build_char_representation(text)
                if self.charmodel_backward_transform is not None:
                    all_backward_chars = [self.charmodel_backward_transform(x) for x in all_backward_chars]
                all_backward_chars = pack(pad_sequence(all_backward_chars, batch_first=True))

                inputs += [all_forward_chars, all_backward_chars]
            else:
                char_reps = self.charmodel(wordchars, wordchars_mask, word_orig_idx, sentlens, wordlens)
                char_reps = PackedSequence(self.trans_char(self.drop(char_reps.data)), char_reps.batch_sizes)
                inputs += [char_reps]

        if self.bert_model is not None:
            device = next(self.parameters()).device
            processed_bert = extract_bert_embeddings(self.args['bert_model'], self.bert_tokenizer, self.bert_model, text, device, keep_endpoints=False,
                                                     num_layers=self.bert_layer_mix.in_features if self.bert_layer_mix is not None else None,
                                                     detach=not self.args.get('bert_finetune', False) or not self.training,
                                                     peft_name=self.peft_name)

            if self.bert_layer_mix is not None:
                # add the average so that the default behavior is to
                # take an average of the N layers, and anything else
                # other than that needs to be learned
                # TODO: refactor this
                processed_bert = [self.bert_layer_mix(feature).squeeze(2) + feature.sum(axis=2) / self.bert_layer_mix.in_features for feature in processed_bert]

            processed_bert = pad_sequence(processed_bert, batch_first=True)
            inputs += [pack(processed_bert)]

        lstm_inputs = torch.cat([x.data for x in inputs], 1)
        lstm_inputs = self.worddrop(lstm_inputs, self.drop_replacement)
        lstm_inputs = self.drop(lstm_inputs)
        lstm_inputs = PackedSequence(lstm_inputs, inputs[0].batch_sizes)

        lstm_outputs, _ = self.taggerlstm(lstm_inputs, sentlens, hx=(self.taggerlstm_h_init.expand(2 * self.args['num_layers'], word.size(0), self.args['hidden_dim']).contiguous(), self.taggerlstm_c_init.expand(2 * self.args['num_layers'], word.size(0), self.args['hidden_dim']).contiguous()))
        lstm_outputs = lstm_outputs.data

        upos_hid = F.relu(self.upos_hid(self.drop(lstm_outputs)))
        upos_pred = self.upos_clf(self.drop(upos_hid))

        preds = [pad(upos_pred).max(2)[1]]

        if upos is not None:
            upos = pack(upos).data
            loss = self.crit(upos_pred.view(-1, upos_pred.size(-1)), upos.view(-1))
        else:
            loss = 0.0

        if self.share_hid:
            xpos_hid = upos_hid
            ufeats_hid = upos_hid

            clffunc = lambda clf, hid: clf(self.drop(hid))
        else:
            xpos_hid = F.relu(self.xpos_hid(self.drop(lstm_outputs)))
            ufeats_hid = F.relu(self.ufeats_hid(self.drop(lstm_outputs)))

            if self.training and upos is not None:
                upos_emb = self.upos_emb(upos)
            else:
                upos_emb = self.upos_emb(upos_pred.max(1)[1])

            clffunc = lambda clf, hid: clf(self.drop(hid), self.drop(upos_emb))

        if xpos is not None: xpos = pack(xpos).data
        if isinstance(self.vocab['xpos'], CompositeVocab):
            xpos_preds = []
            for i in range(len(self.vocab['xpos'])):
                xpos_pred = clffunc(self.xpos_clf[i], xpos_hid)
                if xpos is not None:
                    loss += self.crit(xpos_pred.view(-1, xpos_pred.size(-1)), xpos[:, i].view(-1))
                xpos_preds.append(pad(xpos_pred).max(2, keepdim=True)[1])
            preds.append(torch.cat(xpos_preds, 2))
        else:
            xpos_pred = clffunc(self.xpos_clf, xpos_hid)
            if xpos is not None:
                loss += self.crit(xpos_pred.view(-1, xpos_pred.size(-1)), xpos.view(-1))
            preds.append(pad(xpos_pred).max(2)[1])

        ufeats_preds = []
        if ufeats is not None: ufeats = pack(ufeats).data
        for i in range(len(self.vocab['feats'])):
            ufeats_pred = clffunc(self.ufeats_clf[i], ufeats_hid)
            if ufeats is not None:
                loss += self.crit(ufeats_pred.view(-1, ufeats_pred.size(-1)), ufeats[:, i].view(-1))
            ufeats_preds.append(pad(ufeats_pred).max(2, keepdim=True)[1])
        preds.append(torch.cat(ufeats_preds, 2))

        return loss, preds



"""
Basic testing of French pipeline

The benefit of this test is to verify that the bulk processing works
for languages with MWT in them
"""

import pytest
import stanza
from stanza.models.common.doc import Document

from stanza.tests import *
from stanza.tests.pipeline.pipeline_device_tests import check_on_gpu, check_on_cpu

pytestmark = pytest.mark.pipeline


FR_MWT_SENTENCE = "Alors encore inconnu du grand public, Emmanuel Macron devient en 2014 ministre de l'Économie, de " \
                  "l'Industrie et du Numérique."

EXPECTED_RESULT = """
[
  [
    {
      "id": 1,
      "text": "Alors",
      "lemma": "alors",
      "upos": "ADV",
      "head": 3,
      "deprel": "advmod",
      "start_char": 0,
      "end_char": 5
    },
    {
      "id": 2,
      "text": "encore",
      "lemma": "encore",
      "upos": "ADV",
      "head": 3,
      "deprel": "advmod",
      "start_char": 6,
      "end_char": 12
    },
    {
      "id": 3,
      "text": "inconnu",
      "lemma": "inconnu",
      "upos": "ADJ",
      "feats": "Gender=Masc|Number=Sing",
      "head": 11,
      "deprel": "advcl",
      "start_char": 13,
      "end_char": 20
    },
    {
      "id": [
        4,
        5
      ],
      "text": "du",
      "start_char": 21,
      "end_char": 23
    },
    {
      "id": 4,
      "text": "de",
      "lemma": "de",
      "upos": "ADP",
      "head": 7,
      "deprel": "case"
    },
    {
      "id": 5,
      "text": "le",
      "lemma": "le",
      "upos": "DET",
      "feats": "Definite=Def|Gender=Masc|Number=Sing|PronType=Art",
      "head": 7,
      "deprel": "det"
    },
    {
      "id": 6,
      "text": "grand",
      "lemma": "grand",
      "upos": "ADJ",
      "feats": "Gender=Masc|Number=Sing",
      "head": 7,
      "deprel": "amod",
      "start_char": 24,
      "end_char": 29
    },
    {
      "id": 7,
      "text": "public",
      "lemma": "public",
      "upos": "NOUN",
      "feats": "Gender=Masc|Number=Sing",
      "head": 3,
      "deprel": "obl:arg",
      "start_char": 30,
      "end_char": 36,
      "misc": "SpaceAfter=No"
    },
    {
      "id": 8,
      "text": ",",
      "lemma": ",",
      "upos": "PUNCT",
      "head": 3,
      "deprel": "punct",
      "start_char": 36,
      "end_char": 37
    },
    {
      "id": 9,
      "text": "Emmanuel",
      "lemma": "Emmanuel",
      "upos": "PROPN",
      "head": 11,
      "deprel": "nsubj",
      "start_char": 38,
      "end_char": 46
    },
    {
      "id": 10,
      "text": "Macron",
      "lemma": "Macron",
      "upos": "PROPN",
      "head": 9,
      "deprel": "flat:name",
      "start_char": 47,
      "end_char": 53
    },
    {
      "id": 11,
      "text": "devient",
      "lemma": "devenir",
      "upos": "VERB",
      "feats": "Mood=Ind|Number=Sing|Person=3|Tense=Pres|VerbForm=Fin",
      "head": 0,
      "deprel": "root",
      "start_char": 54,
      "end_char": 61
    },
    {
      "id": 12,
      "text": "en",
      "lemma": "en",
      "upos": "ADP",
      "head": 13,
      "deprel": "case",
      "start_char": 62,
      "end_char": 64
    },
    {
      "id": 13,
      "text": "2014",
      "lemma": "2014",
      "upos": "NUM",
      "feats": "Number=Plur",
      "head": 11,
      "deprel": "obl:mod",
      "start_char": 65,
      "end_char": 69
    },
    {
      "id": 14,
      "text": "ministre",
      "lemma": "ministre",
      "upos": "NOUN",
      "feats": "Gender=Masc|Number=Sing",
      "head": 11,
      "deprel": "xcomp",
      "start_char": 70,
      "end_char": 78
    },
    {
      "id": 15,
      "text": "de",
      "lemma": "de",
      "upos": "ADP",
      "head": 17,
      "deprel": "case",
      "start_char": 79,
      "end_char": 81
    },
    {
      "id": 16,
      "text": "l'",
      "lemma": "le",
      "upos": "DET",
      "feats": "Definite=Def|Number=Sing|PronType=Art",
      "head": 17,
      "deprel": "det",
      "start_char": 82,
      "end_char": 84,
      "misc": "SpaceAfter=No"
    },
    {
      "id": 17,
      "text": "Économie",
      "lemma": "économie",
      "upos": "NOUN",
      "feats": "Gender=Fem|Number=Sing",
      "head": 14,
      "deprel": "nmod",
      "start_char": 84,
      "end_char": 92,
      "misc": "SpaceAfter=No"
    },
    {
      "id": 18,
      "text": ",",
      "lemma": ",",
      "upos": "PUNCT",
      "head": 21,
      "deprel": "punct",
      "start_char": 92,
      "end_char": 93
    },
    {
      "id": 19,
      "text": "de",
      "lemma": "de",
      "upos": "ADP",
      "head": 21,
      "deprel": "case",
      "start_char": 94,
      "end_char": 96
    },
    {
      "id": 20,
      "text": "l'",
      "lemma": "le",
      "upos": "DET",
      "feats": "Definite=Def|Number=Sing|PronType=Art",
      "head": 21,
      "deprel": "det",
      "start_char": 97,
      "end_char": 99,
      "misc": "SpaceAfter=No"
    },
    {
      "id": 21,
      "text": "Industrie",
      "lemma": "industrie",
      "upos": "NOUN",
      "feats": "Gender=Fem|Number=Sing",
      "head": 17,
      "deprel": "conj",
      "start_char": 99,
      "end_char": 108
    },
    {
      "id": 22,
      "text": "et",
      "lemma": "et",
      "upos": "CCONJ",
      "head": 25,
      "deprel": "cc",
      "start_char": 109,
      "end_char": 111
    },
    {
      "id": [
        23,
        24
      ],
      "text": "du",
      "start_char": 112,
      "end_char": 114
    },
    {
      "id": 23,
      "text": "de",
      "lemma": "de",
      "upos": "ADP",
      "head": 25,
      "deprel": "case"
    },
    {
      "id": 24,
      "text": "le",
      "lemma": "le",
      "upos": "DET",
      "feats": "Definite=Def|Gender=Masc|Number=Sing|PronType=Art",
      "head": 25,
      "deprel": "det"
    },
    {
      "id": 25,
      "text": "Numérique",
      "lemma": "numérique",
      "upos": "NOUN",
      "feats": "Gender=Masc|Number=Sing",
      "head": 17,
      "deprel": "conj",
      "start_char": 115,
      "end_char": 124,
      "misc": "SpaceAfter=No"
    },
    {
      "id": 26,
      "text": ".",
      "lemma": ".",
      "upos": "PUNCT",
      "head": 11,
      "deprel": "punct",
      "start_char": 124,
      "end_char": 125,
      "misc": "SpaceAfter=No"
    }
  ]
]
"""

class TestFrenchPipeline:
    @pytest.fixture(scope="class")
    def pipeline(self):
        """ Create a pipeline with French models """
        pipeline = stanza.Pipeline(processors='tokenize,mwt,pos,lemma,depparse', dir=TEST_MODELS_DIR, lang='fr')
        return pipeline

    def test_single(self, pipeline):
        doc = pipeline(FR_MWT_SENTENCE)
        compare_ignoring_whitespace(str(doc), EXPECTED_RESULT)

    def test_bulk(self, pipeline):
        NUM_DOCS = 10
        raw_text = [FR_MWT_SENTENCE] * NUM_DOCS
        raw_doc = [Document([], text=doccontent) for doccontent in raw_text]

        result = pipeline(raw_doc)

        assert len(result) == NUM_DOCS
        for doc in result:
            compare_ignoring_whitespace(str(doc), EXPECTED_RESULT)
            assert len(doc.sentences) == 1
            assert doc.num_words == 26
            assert doc.num_tokens == 24

    def test_on_gpu(self, pipeline):
        """
        The default pipeline should have all the models on the GPU
        """
        check_on_gpu(pipeline)

    def test_on_cpu(self):
        """
        Create a pipeline on the CPU, check that all the models on CPU
        """
        pipeline = stanza.Pipeline("fr", dir=TEST_MODELS_DIR, use_gpu=False)
        check_on_cpu(pipeline)



from enum import Enum
import logging
import os

import torch

from pickle import UnpicklingError
import warnings

logger = logging.getLogger('stanza')

class ModelType(Enum):
    LSTM               = 1
    ENSEMBLE           = 2

class BaseTrainer:
    def __init__(self, model, optimizer=None, scheduler=None, epochs_trained=0, batches_trained=0, best_f1=0.0, best_epoch=0, first_optimizer=False):
        self.model = model
        self.optimizer = optimizer
        self.scheduler = scheduler
        # keeping track of the epochs trained will be useful
        # for adjusting the learning scheme
        self.epochs_trained = epochs_trained
        self.batches_trained = batches_trained
        self.best_f1 = best_f1
        self.best_epoch = best_epoch
        self.first_optimizer = first_optimizer

    def save(self, filename, save_optimizer=True):
        params = self.model.get_params()
        checkpoint = {
            'params': params,
            'epochs_trained': self.epochs_trained,
            'batches_trained': self.batches_trained,
            'best_f1': self.best_f1,
            'best_epoch': self.best_epoch,
            'model_type': self.model_type.name,
            'first_optimizer': self.first_optimizer,
        }
        checkpoint["bert_lora"] = self.get_peft_params()
        if save_optimizer and self.optimizer is not None:
            checkpoint['optimizer_state_dict'] = self.optimizer.state_dict()
            checkpoint['scheduler_state_dict'] = self.scheduler.state_dict()
        torch.save(checkpoint, filename, _use_new_zipfile_serialization=False)
        logger.info("Model saved to %s", filename)

    def log_norms(self):
        self.model.log_norms()

    def log_shapes(self):
        self.model.log_shapes()

    @property
    def transitions(self):
        return self.model.transitions

    @property
    def root_labels(self):
        return self.model.root_labels

    @property
    def device(self):
        return next(self.model.parameters()).device

    def train(self):
        return self.model.train()

    def eval(self):
        return self.model.eval()

    # TODO: make ABC with methods such as model_from_params?
    # TODO: if we save the type in the checkpoint, use that here to figure out which to load
    @staticmethod
    def load(filename, args=None, load_optimizer=False, foundation_cache=None, peft_name=None):
        """
        Load back a model and possibly its optimizer.
        """
        # hide the import here to avoid circular imports
        from stanza.models.constituency.ensemble import EnsembleTrainer
        from stanza.models.constituency.trainer import Trainer

        if not os.path.exists(filename):
            if args.get('save_dir', None) is None:
                raise FileNotFoundError("Cannot find model in {} and args['save_dir'] is None".format(filename))
            elif os.path.exists(os.path.join(args['save_dir'], filename)):
                filename = os.path.join(args['save_dir'], filename)
            else:
                raise FileNotFoundError("Cannot find model in {} or in {}".format(filename, os.path.join(args['save_dir'], filename)))
        try:
            # TODO: currently cannot switch this to weights_only=True
            # without in some way changing the model to save enums in
            # a safe manner, probably by converting to int
            try:
                checkpoint = torch.load(filename, lambda storage, loc: storage, weights_only=True)
            except UnpicklingError as e:
                checkpoint = torch.load(filename, lambda storage, loc: storage, weights_only=False)
                warnings.warn("The saved constituency parser has an old format using Enum, set, unsanitized Transitions, etc.  This version of Stanza can support reading both the new and the old formats.  Future versions will only allow loading with weights_only=True.  Please resave the constituency parser using this version ASAP.")
        except BaseException:
            logger.exception("Cannot load model from %s", filename)
            raise
        logger.debug("Loaded model from %s", filename)

        params = checkpoint['params']

        if 'model_type' not in checkpoint:
            # old models will have this trait
            # TODO: can remove this after 1.10
            checkpoint['model_type'] = ModelType.LSTM
        if isinstance(checkpoint['model_type'], str):
            checkpoint['model_type'] = ModelType[checkpoint['model_type']]
        if checkpoint['model_type'] == ModelType.LSTM:
            clazz = Trainer
        elif checkpoint['model_type'] == ModelType.ENSEMBLE:
            clazz = EnsembleTrainer
        else:
            raise ValueError("Unexpected model type: %s" % checkpoint['model_type'])
        model = clazz.model_from_params(params, checkpoint.get('bert_lora', None), args, foundation_cache, peft_name)

        epochs_trained = checkpoint['epochs_trained']
        batches_trained = checkpoint.get('batches_trained', 0)
        best_f1 = checkpoint['best_f1']
        best_epoch = checkpoint['best_epoch']

        if 'first_optimizer' not in checkpoint:
            # this will only apply to old (LSTM) Trainers
            # EnsembleTrainers will always have this value saved
            # so here we can compensate by looking at the old training statistics...
            # we use params['config'] here instead of model.args
            # because the args might have a different training
            # mechanism, but in order to reload the optimizer, we need
            # to match the optimizer we build with the one that was
            # used at training time
            build_simple_adadelta = params['config']['multistage'] and epochs_trained < params['config']['epochs'] // 2
            checkpoint['first_optimizer'] = build_simple_adadelta
        first_optimizer = checkpoint['first_optimizer']

        if load_optimizer:
            optimizer = clazz.load_optimizer(model, checkpoint, first_optimizer, filename)
            scheduler = clazz.load_scheduler(model, optimizer, checkpoint, first_optimizer)
        else:
            optimizer = None
            scheduler = None

        if checkpoint['model_type'] == ModelType.LSTM:
            logger.debug("-- MODEL CONFIG --")
            for k in model.args.keys():
                logger.debug("  --%s: %s", k, model.args[k])
            return Trainer(model=model, optimizer=optimizer, scheduler=scheduler, epochs_trained=epochs_trained, batches_trained=batches_trained, best_f1=best_f1, best_epoch=best_epoch, first_optimizer=first_optimizer)
        elif checkpoint['model_type'] == ModelType.ENSEMBLE:
            return EnsembleTrainer(ensemble=model, optimizer=optimizer, scheduler=scheduler, epochs_trained=epochs_trained, batches_trained=batches_trained, best_f1=best_f1, best_epoch=best_epoch, first_optimizer=first_optimizer)
        else:
            raise ValueError("Unexpected model type: %s" % checkpoint['model_type'])

